"Functions specific to GenX outputs"

import logging
from dataclasses import dataclass
from itertools import product
from pathlib import Path
from typing import Dict, List

import pandas as pd

from powergenome.external_data import (
    load_demand_segments,
    load_policy_scenarios,
    make_generator_variability,
)
from powergenome.financials import investment_cost_calculator
from powergenome.time_reduction import kmeans_time_clustering
from powergenome.util import find_region_col, snake_case_col, snake_case_str

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class FolderStructure:
    resource: Path
    policy_assignments: Path
    system: Path
    policy: Path


@dataclass(frozen=True, slots=True)
class GenXInputData:
    """A class to abstract the dataframes that will be written out in the GenX case folder."""

    tag: str
    folder: Path
    file_name: str
    dataframe: pd.DataFrame

    def __post_init__(self):
        """Validate the data after initialization."""
        if not isinstance(self.folder, Path):
            object.__setattr__(self, "folder", Path(self.folder))
        if not self.file_name.endswith(".csv"):
            raise ValueError(f"file_name must end with .csv, got {self.file_name}")
        if not isinstance(self.dataframe, (pd.DataFrame, type(None))):
            raise TypeError(
                f"dataframe must be a pandas DataFrame or None, got {type(self.dataframe)}"
            )


INT_COLS = [
    "Inv_Cost_per_MWyr",
    "Fixed_OM_Cost_per_MWyr",
    "Inv_Cost_per_MWhyr",
    "Fixed_OM_Cost_per_MWhyr",
    "Line_Reinforcement_Cost_per_MW_yr",
    "Up_Time",
    "Down_Time",
]

COL_ROUND_VALUES = {
    "Var_OM_Cost_per_MWh": 2,
    "Var_OM_Cost_per_MWh_in": 2,
    "Start_Cost_per_MW": 0,
    "Cost_per_MMBtu": 2,
    "CO2_content_tons_per_MMBtu": 5,
    "Cap_Size": 2,
    "Heat_Rate_MMBTU_per_MWh": 2,
    "distance_mile": 4,
    "Line_Max_Reinforcement_MW": 0,
    "distance_miles": 1,
    "distance_km": 1,
    "Existing_Cap_MW": 1,
    "Existing_Cap_MWh": 1,
    "Max_Cap_MW": 1,
    "Max_Cap_MWh": 1,
    "Min_Cap_MW": 1,
    "Min_Cap_MWh": 1,
    "Line_Loss_Percentage": 4,
}

# RESOURCE_TAGS = ["THERM", "VRE", "MUST_RUN", "STOR", "FLEX", "HYDRO", "LDS"]
RESOURCE_TAGS = ["THERM", "VRE", "MUST_RUN", "STOR", "FLEX", "HYDRO"]

DEFAULT_COLS = ["Resource", "Zone", "New_Build", "Can_Retire", "Existing_Cap_MW"]

# Specific columns for each resource type
THERM_COLUMNS = [
    "Down_Time",
    "Min_Power",
    "Ramp_Up_Percentage",
    "Ramp_Dn_Percentage",
    "Up_Time",
]

VRE_COLUMNS = ["Num_VRE_Bins"]

MUST_RUN_COLUMNS = []

STOR_COLUMNS = [
    "Eff_Down",
    "Eff_Up",
    "Existing_Cap_MWh",
    "Existing_Charge_Cap_MW",
    "Fixed_OM_Cost_Charge_per_MWyr",
    "Fixed_OM_Cost_per_MWhyr",
    "Inv_Cost_Charge_per_MWyr",
    "Inv_Cost_per_MWhyr",
    "LDS",
    "Max_Cap_MWh",
    "Max_Charge_Cap_MW",
    "Max_Duration",
    "Min_Cap_MWh",
    "Min_Charge_Cap_MW",
    "Min_Duration",
    "Self_Disch",
    "Var_OM_Cost_per_MWh_In",
]

FLEX_COLUMNS = [
    "Flexible_Demand_Energy_Eff",
    "Max_Flexible_Demand_Delay",
    "Max_Flexible_Demand_Advance",
    "Var_OM_Cost_per_MWh_In",
]

HYDRO_COLUMNS = [
    "Eff_Up",
    "Eff_Down",
    "Hydro_Energy_to_Power_Ratio",
    "LDS",
    "Min_Power",
    "Ramp_Up_Percentage",
    "Ramp_Dn_Percentage",
]

# For future use
# HYDROGEN_COLUMNS = [
#     "Hydrogen_MWh_Per_Tonne",
#     "Hydrogen_Price_Per_Tonne",
#     "Min_Power",
#     "Ramp_Up_Percentage",
#     "Ramp_Dn_Percentage",
# ]

# Create a mapping of resource tags -> columns
RESOURCE_COLUMNS = {
    "THERM": THERM_COLUMNS,
    "VRE": VRE_COLUMNS,
    "MUST_RUN": MUST_RUN_COLUMNS,
    "STOR": STOR_COLUMNS,
    "FLEX": FLEX_COLUMNS,
    "HYDRO": HYDRO_COLUMNS,
    # 'ELECTROLYZER': HYDROGEN_COLUMNS,
}

# Create a mapping of resource tags -> filenames
RESOURCE_FILENAMES = {
    "THERM": "Thermal.csv",
    "VRE": "Vre.csv",
    "MUST_RUN": "Must_run.csv",
    "STOR": "Storage.csv",
    "FLEX": "Flex_demand.csv",
    "HYDRO": "Hydro.csv",
    # "ELECTROLYZER": "Electrolyzer.csv",
}

MULTISTAGE_COLS = [
    "WACC",
    "Capital_Recovery_Period",
    "Lifetime",
    "Min_Retired_Cap_MW",
    "Min_Retired_Energy_Cap_MW",
    "Min_Retired_Charge_Cap_MW",
]

# Create a mapping to update the policy tag columns
POLICY_TAGS = [
    ("ESR", {"oldtag": "ESR_", "newtag": "ESR_"}),
    ("CAP_RES", {"oldtag": "CapRes_", "newtag": "Derating_factor_"}),
    ("MIN_CAP", {"oldtag": "MinCapTag_", "newtag": "Min_Cap_"}),
    ("MAX_CAP", {"oldtag": "MaxCapTag_", "newtag": "Max_Cap_"}),
]

POLICY_TAGS_FILENAMES = {
    "ESR": "Resource_energy_share_requirement.csv",
    "CAP_RES": "Resource_capacity_reserve_margin.csv",
    "MIN_CAP": "Resource_minimum_capacity_requirement.csv",
    "MAX_CAP": "Resource_maximum_capacity_requirement.csv",
}


def create_policy_req(settings: dict, col_str_match: str) -> pd.DataFrame:
    model_year = settings["model_year"]
    case_id = settings["case_id"]

    policies = load_policy_scenarios(settings)
    policy_cols = [c for c in policies.columns if col_str_match in c]
    if len(policy_cols) == 0:
        return None

    year_case_policy = policies.loc[(case_id, model_year), ["region"] + policy_cols]
    if isinstance(year_case_policy, pd.DataFrame):
        year_case_policy = year_case_policy.dropna(subset=policy_cols)
    elif isinstance(year_case_policy, pd.Series):
        year_case_policy = year_case_policy.dropna()
    else:
        raise TypeError(
            "Somehow the object 'year_case_policy' is not a dataframe of a series."
            "Please submit this as an issue in the PowerGenome repository."
            f"{year_case_policy}"
        )
    # Bug where multiple regions for a case will return this as a df, even if the policy
    # for this case applies to all regions (code below expects a Series)
    ycp_shape = year_case_policy.shape
    if ycp_shape[0] == 1 and len(ycp_shape) > 1:
        year_case_policy = year_case_policy.squeeze()  # convert to series

    zones = settings["model_regions"]
    zone_num_map = {
        zone: f"z{number + 1}" for zone, number in zip(zones, range(len(zones)))
    }

    zone_cols = ["Region_description", "Network_zones"] + policy_cols
    zone_df = pd.DataFrame(columns=zone_cols, dtype=float)
    zone_df["Region_description"] = zones
    zone_df["Network_zones"] = zone_df["Region_description"].map(zone_num_map)
    # If there is only one region, assume that the policy is applied across all regions.
    if isinstance(year_case_policy, pd.Series):
        logger.info(
            "Only one zone was found in the emissions policy file."
            " The same emission policies are being applied to all zones."
        )
        for col, value in year_case_policy[policy_cols].iteritems():
            if "CO_2_Max_Mtons" in col:
                zone_df.loc[:, col] = 0
                if value > 0:
                    zone_df.loc[0, col] = value
            else:
                zone_df.loc[:, col] = value
    else:
        for region, col in product(
            year_case_policy["region"].unique(), year_case_policy[policy_cols].columns
        ):
            zone_df.loc[zone_df["Region_description"] == region, col] = (
                year_case_policy.loc[year_case_policy.region == region, col].values[0]
            )

    # zone_df = zone_df.drop(columns="region")

    return zone_df


def create_regional_cap_res(settings: dict) -> pd.DataFrame:
    """Create a dataframe of regional capacity reserve constraints from settings params

    Parameters
    ----------
    settings : dict
        PowerGenome settings dictionary with parameters 'model_regions' and
        'regional_capacity_reserves'

    Returns
    -------
    pd.DataFrame
        A dataframe with a 'Network_zones' column and one 'CapRes_*' for each capacity
        reserve constraint

    Raises
    ------
    KeyError
        A region listed in a capacity reserve constraint is not a valid model region
    """

    if not settings.get("regional_capacity_reserves"):
        return None
    else:
        zones = settings["model_regions"]
        zone_num_map = {
            zone: f"z{number + 1}" for zone, number in zip(zones, range(len(zones)))
        }
        cap_res_cols = list(settings["regional_capacity_reserves"])
        cap_res_df = pd.DataFrame(index=zones, columns=["Network_zones"] + cap_res_cols)
        cap_res_df["Network_zones"] = cap_res_df.index.map(zone_num_map)

        for col, cap_res_dict in settings["regional_capacity_reserves"].items():
            for region, val in cap_res_dict.items():
                if region not in zones:
                    raise KeyError(
                        f"The region {region} in 'regional_capacity_reserves', {col} "
                        "is not a valid model region."
                    )
                cap_res_df.loc[region, col] = val

        cap_res_df = cap_res_df.fillna(0)

        return cap_res_df


def label_cap_res_lines(path_names: List[str], dest_regions: List[str]) -> List[int]:
    """Label if each transmission line is part of a capacity reserve constraint and
    if line flow is into or out of the constraint region.

    Parameters
    ----------
    path_names : List[str]
        String names of transmission lines, with format <region>_to_<region>
    dest_regions : List[str]
        Names of model regions, corresponding to region names used in 'path_names'

    Returns
    -------
    List[int]
        Same length as 'path_names'. Values of 1 mean the line connects a region within
        a capacity reserve constraint to a region outside the constraint. Values of -1
        mean it connects a region outside a capacity reserve constraint to a region
        within the constraint. Values of 0 mean it connects two regions that are both
        within or outside the constraint.
    """
    cap_res_list = []
    for name in path_names:
        s_r = name.split("_to_")[0]
        e_r = name.split("_to_")[-1]
        if (s_r in dest_regions) and e_r not in dest_regions:
            cap_res_list.append(1)
        elif (e_r in dest_regions) and s_r not in dest_regions:
            cap_res_list.append(-1)
        else:
            cap_res_list.append(0)

    assert len(cap_res_list) == len(path_names)

    return cap_res_list


def add_cap_res_network(tx_df: pd.DataFrame, settings: dict) -> pd.DataFrame:
    """Add capacity reserve colums to the transmission dataframe (Network.csv)

    Parameters
    ----------
    tx_df : pd.DataFrame
        Transmission dataframe with rows for each transmission line, a column
        'transmission_path_name', and columns 'z1', 'z2', etc for each model region. The
        'z*' columns have values of -1 (line is outbound from region), 0 (not connected),
        or 1 (inbound to region).
    settings : dict
        PowerGenome settings, with parameters 'model_regions', 'regional_capacity_reserves',
        and 'cap_res_network_derate_default'.


    Returns
    -------
    pd.DataFrame
        A copy of the input dataframe with additional columns 'CapRes_*', 'DerateCapRes_*',
        and 'CapRes_Excl_*' for each capacity reserve constraint
    """
    if (
        "Transmission Path Name" in tx_df.columns
        and "transmission_path_name" not in tx_df.columns
    ):
        tx_df["transmission_path_name"] = tx_df["Transmission Path Name"]
    original_cols = tx_df.columns.to_list()

    path_names = tx_df["transmission_path_name"].to_list()
    policy_nums = []

    # Loop through capacity reserve constraints (CapRes_*) and determine network
    # parameters for each
    for cap_res in settings.get("regional_capacity_reserves", {}) or {}:
        cap_res_num = int(cap_res.split("_")[-1])  # the number of the capres constraint
        policy_nums.append(cap_res_num)
        # TODO #179 fix reference to regional_capacity_reserves key of settings dict
        dest_regions = list(
            settings["regional_capacity_reserves"][cap_res].keys()
        )  # list of regions in the CapRes
        # May add ability to have different values by CapRes and line in the future
        tx_df[f"DerateCapRes_{cap_res_num}"] = settings.get(
            "cap_res_network_derate_default", 0.95
        )
        tx_df[f"CapRes_Excl_{cap_res_num}"] = label_cap_res_lines(
            path_names, dest_regions
        )

    policy_nums.sort()
    derate_cols = [f"DerateCapRes_{n}" for n in policy_nums]
    excl_cols = [f"CapRes_Excl_{n}" for n in policy_nums]

    return tx_df[original_cols + derate_cols + excl_cols].fillna(0)


def add_emission_policies(transmission_df, settings):
    """Add emission policies to the transmission dataframe

    Parameters
    ----------
    transmission_df : DataFrame
        Zone to zone transmission constraints
    settings : dict
        User-defined parameters from a settings file. Should have keys of `input_folder`
        (a Path object of where to find user-supplied data) and
        `emission_policies_fn` (the file to load).
    DistrZones : [type], optional
        Placeholder setting, by default None

    Returns
    -------
    DataFrame
        The emission policies provided by user next to the transmission constraints.
    """

    model_year = settings["model_year"]
    case_id = settings["case_id"]

    policies = load_policy_scenarios(settings)
    year_case_policy = policies.loc[(case_id, model_year), :]

    # Bug where multiple regions for a case will return this as a df, even if the policy
    # for this case applies to all regions (code below expects a Series)
    ycp_shape = year_case_policy.shape
    if ycp_shape[0] == 1 and len(ycp_shape) > 1:
        year_case_policy = year_case_policy.squeeze()  # convert to series

    zones = settings["model_regions"]
    zone_num_map = {
        zone: f"z{number + 1}" for zone, number in zip(zones, range(len(zones)))
    }

    zone_cols = ["Region description", "Network_zones"] + list(policies.columns)
    zone_df = pd.DataFrame(columns=zone_cols)
    zone_df["Region description"] = zones
    zone_df["Network_zones"] = zone_df["Region description"].map(zone_num_map)

    # Add code here to make DistrZones something else!
    # If there is only one region, assume that the policy is applied across all regions.
    if isinstance(year_case_policy, pd.Series):
        logger.info(
            "Only one zone was found in the emissions policy file."
            " The same emission policies are being applied to all zones."
        )
        for col, value in year_case_policy.iteritems():
            if col == "CO_2_Max_Mtons":
                zone_df.loc[:, col] = 0
                if value > 0:
                    zone_df.loc[0, col] = value
            else:
                zone_df.loc[:, col] = value
    else:
        for region, col in product(
            year_case_policy["region"].unique(), year_case_policy.columns
        ):
            zone_df.loc[zone_df["Region description"] == region, col] = (
                year_case_policy.loc[year_case_policy.region == region, col].values[0]
            )

    zone_df = zone_df.drop(columns="region")

    network_df = pd.concat([zone_df, transmission_df], axis=1)

    return network_df


def add_misc_gen_values(
    gen_clusters: pd.DataFrame, settings: dict, resource_col: str = "Resource"
) -> pd.DataFrame:
    """Add parameter values from a CSV file to resources.

    Parameters
    ----------
    gen_clusters : pd.DataFrame
        Resource dataframe with columns "region" and `resource_col`.
    settings : dict
        Model settings, with parameters "input_folder" and "misc_gen_inputs_fn". The
        misc gen CSV file should have the column `resource_col`. If it has the column
        "region" then regional values will be applied, otherwise values for each resource
        will be applied across all regions.
    resource_col : str, optional
        Name of the column with resource name in both gen_clusters and the CSV file, by
        default "Resource".

    Returns
    -------
    pd.DataFrame
        A modified version of gen_clusters with new parameter values for resources.
    """
    path = Path(settings["input_folder"]) / settings["misc_gen_inputs_fn"]
    misc_values = pd.read_csv(path)
    misc_values[resource_col] = snake_case_col(misc_values[resource_col])

    context = f"Assigning misc generator values from the user-supplied file {path}."
    try:
        region_col = find_region_col(misc_values.columns, context)
    except ValueError:
        region_col = "region"
        misc_values["region"] = "all"
    regions = [
        r for r in misc_values[region_col].fillna("all").unique() if r.lower() != "all"
    ]
    wrong_regions = [r for r in regions if r not in settings["model_regions"]]
    if wrong_regions:
        raise ValueError(
            f"The `misc_gen_inputs_fn` CSV has regions {wrong_regions}, which are not "
            f"valid model regions. Valid model regions are {settings['model_regions']}."
        )

    for region in settings["model_regions"]:
        _df = misc_values.loc[misc_values[region_col].str.lower() == "all", :]
        _df.loc[:, "region"] = region
        misc_values = misc_values.append(_df)

    misc_values = misc_values.loc[misc_values[region_col].str.lower() != "all", :]

    for tech, _df in misc_values.groupby(resource_col):
        num_tech_regions = len(
            gen_clusters.loc[
                gen_clusters[resource_col].str.contains(tech, case=False, regex=False)
            ].drop_duplicates(subset=["region"])
        )
        num_values = len(_df)
        if num_values < num_tech_regions:
            logger.warning(
                f"The `misc_gen_inputs_fn` CSV has {num_values} region(s) for the resource "
                f"'{tech}', but the resource is in {num_tech_regions} regions. Check "
                "your input file to ensure values are provided for all appropriate regions."
            )
    generic_resources = []
    for gen_resource in gen_clusters[resource_col].unique():
        for r in sorted(settings["model_regions"])[::-1]:
            if r in gen_resource:
                gen_resource = gen_resource.replace(r + "_", "")
                generic_resources.append(snake_case_str(gen_resource))
                continue
    generic_resources = set(generic_resources)
    missing_resources = []
    for resource in generic_resources:
        match = False
        for misc_resource in misc_values[resource_col].unique():
            if misc_resource in resource.lower():
                match = True
                continue
        if not match:
            missing_resources.append(resource)

    if missing_resources:
        logger.warning(
            f"The resources {missing_resources} are not included in your `misc_gen_inputs_fn` "
            "CSV file. This is a warning in case they should have parameters in that file."
        )

    misc_values = misc_values.reset_index(drop=True)
    value_cols = [
        col for col in misc_values.columns if col not in [region_col, resource_col]
    ]

    for idx, row in misc_values.iterrows():
        row_cols = row[value_cols].dropna().index
        gen_clusters.loc[
            (gen_clusters["region"] == row[region_col])
            & (
                snake_case_col(gen_clusters[resource_col]).str.contains(
                    row[resource_col], case=False
                )
            ),
            row_cols,
        ] = row[row_cols].values
    return gen_clusters


def reduce_time_domain(
    resource_profiles, load_profiles, settings, variable_resources_only=True
):
    demand_segments = load_demand_segments(settings)

    if settings.get("reduce_time_domain"):
        days = settings["time_domain_days_per_period"]
        time_periods = settings["time_domain_periods"]
        include_peak_day = settings["include_peak_day"]
        load_weight = settings["demand_weight_factor"]

        results, representative_point, _ = kmeans_time_clustering(
            resource_profiles=resource_profiles,
            load_profiles=load_profiles,
            days_in_group=days,
            num_clusters=time_periods,
            include_peak_day=include_peak_day,
            load_weight=load_weight,
            variable_resources_only=variable_resources_only,
            n_init=settings.get("tdr_n_init", 100),
        )

        reduced_resource_profile = results["resource_profiles"]
        reduced_resource_profile.index.name = "Resource"
        reduced_resource_profile.index = range(1, len(reduced_resource_profile) + 1)
        reduced_load_profile = results["load_profiles"]
        time_series_mapping = results["time_series_mapping"]

        time_index = pd.Series(data=reduced_load_profile.index + 1, name="Time_Index")
        sub_weights = pd.Series(
            data=[x * (days * 24) for x in results["ClusterWeights"]],
            name="Sub_Weights",
        )
        hours_per_period = pd.Series(data=[days * 24], name="Timesteps_per_Rep_Period")
        subperiods = pd.Series(data=[time_periods], name="Rep_Periods")
        reduced_load_output = pd.concat(
            [
                demand_segments,
                subperiods,
                hours_per_period,
                sub_weights,
                time_index,
                reduced_load_profile.round(0),
            ],
            axis=1,
        )

        return (
            reduced_resource_profile,
            reduced_load_output,
            time_series_mapping,
            representative_point,
        )

    else:
        time_index = pd.Series(data=range(1, len(load_profiles) + 1), name="Time_Index")
        sub_weights = pd.Series(data=[len(time_index)], name="Sub_Weights")
        hours_per_period = pd.Series(
            data=[len(load_profiles)], name="Timesteps_per_Rep_Period"
        )
        subperiods = pd.Series(data=[1], name="Rep_Periods")

        # Not actually reduced
        load_output = pd.concat(
            [
                demand_segments,
                subperiods,
                hours_per_period,
                sub_weights,
                time_index,
                load_profiles.reset_index(drop=True),
            ],
            axis=1,
        )
        resource_profiles.index = time_index

        return resource_profiles, load_output, None, None


def network_line_loss(transmission: pd.DataFrame, settings: dict) -> pd.DataFrame:
    """Add line loss percentage for each network line between regions.

    Parameters
    ----------
    transmission : pd.DataFrame
        One network line per row with a column "distance_mile"
    settings : dict
        User-defined settings with a parameter "tx_line_loss_100_miles"

    Returns
    -------
    pd.DataFrame
        Same as input but with the new column 'Line_Loss_Percentage'
    """
    if "tx_line_loss_100_miles" not in settings:
        raise KeyError(
            "The parameter 'tx_line_loss_100_miles' is required in your settings file."
        )
    loss_per_100_miles = settings["tx_line_loss_100_miles"]
    if "distance_mile" in transmission.columns:
        distance_col = "distance_mile"
    elif "distance_km" in transmission.columns:
        distance_col = "distance_km"
        loss_per_100_miles *= 0.62137
        logger.debug("Line loss per 100 miles was converted to km.")
    else:
        raise KeyError("No distance column is available in the transmission dataframe")

    transmission["Line_Loss_Percentage"] = (
        transmission[distance_col] / 100 * loss_per_100_miles
    )

    return transmission


def network_reinforcement_cost(
    transmission: pd.DataFrame, settings: dict
) -> pd.DataFrame:
    """Add transmission line reinforcement investment costs (per MW-mile-year)

    Parameters
    ----------
    transmission : pd.DataFrame
        One network line per row with columns "transmission_path_name" and
        "distance_mile".
    settings : dict
        User-defined settings with dictionary "transmission_investment_cost.tx" with
        keys "capex_mw_mile", "wacc", and "investment_years".

    Returns
    -------
    pd.DataFrame
        Same as input but with the new column 'Line_Reinforcement_Cost_per_MW_yr'
    """

    cost_dict = (
        settings.get("transmission_investment_cost", {})
        .get("tx", {})
        .get("capex_mw_mile")
    )
    if not cost_dict:
        raise KeyError(
            "No value for transmission reinforcement costs is included in the settings."
            " These costs are included under transmission_investment_costs.tx."
            "capex_mw_mile.<model_region>. See the `test_settings.yml` file for an "
            "example."
        )
    origin_region_cost = (
        transmission["transmission_path_name"].str.split("_to_").str[0].map(cost_dict)
    )
    dest_region_cost = (
        transmission["transmission_path_name"].str.split("_to_").str[-1].map(cost_dict)
    )

    # Average the costs per mile between origin and destination regions
    line_capex = (origin_region_cost + dest_region_cost) / 2
    line_wacc = (
        settings.get("transmission_investment_cost", {}).get("tx", {}).get("wacc")
    )
    if not line_wacc:
        raise KeyError(
            "No value for the transmission weighted average cost of capital (wacc) is "
            "included in the settings."
            "This numeric value is included under transmission_investment_costs.tx."
            "wacc. See the `test_settings.yml` file for an "
            "example."
        )
    line_inv_period = (
        settings.get("transmission_investment_cost", {})
        .get("tx", {})
        .get("investment_years")
    )
    if not line_inv_period:
        raise KeyError(
            "No value for the transmission investment period is included in the settings."
            "This numeric value is included under transmission_investment_costs.tx."
            "investment_years. See the `test_settings.yml` file for an example."
        )
    line_inv_cost = (
        investment_cost_calculator(
            line_capex,
            line_wacc,
            line_inv_period,
            settings.get("interest_compound_method", "discrete"),
        )
        * transmission["distance_mile"]
    )

    transmission["Line_Reinforcement_Cost_per_MWyr"] = line_inv_cost.round(0)

    return transmission


def network_max_reinforcement(
    transmission: pd.DataFrame, settings: dict
) -> pd.DataFrame:
    """Add the maximum amount that transmission lines between regions can be reinforced
    in a planning period.

    Parameters
    ----------
    transmission : pd.DataFrame
        One network line per row with the column "Line_Max_Flow_MW"
    settings : dict
        User-defined settings with the parameter "Line_Max_Reinforcement_MW"

    Returns
    -------
    pd.DataFrame
        A copy of the input transmission constraint dataframe with a new column
        `Line_Max_Reinforcement_MW`.
    """

    max_expansion = settings.get("tx_expansion_per_period", 0)
    expansion_mw = settings.get("tx_expansion_mw_per_period", 0)

    if not max_expansion and max_expansion != 0:
        raise KeyError(
            "No value for the transmission expansion allowed in this model period is "
            "included in the settings."
            "This numeric value is included under tx_expansion_per_period. See the "
            "`test_settings.yml` file for an example."
        )
    # if isinstance(max_expansion, dict):
    #     expansion_method = settings.get("tx_expansion_method")
    #     if not expansion_method:
    #         raise KeyError(
    #         "The transmission expansion parameter 'tx_expansion_per_period' is a "
    #         "dictionary. There should also be a settings parameter 'tx_expansion_method' "
    #         "with a dictionary of model region: <type> (either 'capacity' or 'fraction' "
    #         "but it isn't in your settings file."
    #     )
    #     for region, value in max_expansion.items():
    #         existing_tx = transmission.query("Region description == @region")["Line_Max_Flow_MW"]
    #         if expansion_method[region].lower() == "capacity":
    #             transmission.loc[transmission["Region description"] == region,
    #             "Line_Max_Reinforcement_MW"] = value
    #         elif expansion_method[region].lower() == "fraction":
    #             transmission.loc[transmission["Region description"] == region,
    #             "Line_Max_Reinforcement_MW"] = value * existing_tx
    #         else:
    #             raise KeyError(
    #             "The transmission expansion method parameter (tx_expansion_method) "
    #             "should have values of 'capacity' or 'fraction' for each model region. "
    #             f"The value provided was '{expansion_method[region]}'."
    #         )

    # else:
    transmission.loc[:, "Line_Max_Reinforcement_MW"] = [
        max(tx * max_expansion, expansion_mw) for tx in transmission["Line_Max_Flow_MW"]
    ]
    transmission["Line_Max_Reinforcement_MW"] = transmission[
        "Line_Max_Reinforcement_MW"
    ].round(0)

    return transmission


def set_int_cols(df: pd.DataFrame, cols: list = None) -> pd.DataFrame:
    """Set values of some dataframe columns to integers.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    cols : list, optional
        Columns to set as integer, by default None. If none, will use
        `powergenome.GenX.INT_COLS`.

    Returns
    -------
    pd.DataFrame
        Input dataframe with some columns set as integer.
    """
    if not cols:
        cols = INT_COLS

    cols = [c for c in cols if c in df.columns]

    for col in cols:
        df[col] = df[col].fillna(0).astype(int)
    return df


def round_col_values(
    df: pd.DataFrame, col_round_val: Dict[str, int] = None
) -> pd.DataFrame:
    """Round values in columns to a specific sigfig.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe.
    col_round_val : Dict, optional
        Dictionary with key values of column labels and integer values of the number of
        sigfigs, by default None.

    Returns
    -------
    pd.DataFrame
        Same dataframe as input but with rounded values in specified columns.
    """
    if not col_round_val:
        col_round_val = COL_ROUND_VALUES

    col_round_val = {k: v for k, v in col_round_val.items() if k in df.columns}

    for col, value in col_round_val.items():
        df[col] = df[col].fillna(0).round(value)
    return df


def calculate_partial_CES_values(gen_clusters, fuels, settings):
    gens = gen_clusters.copy()
    if settings.get("partial_ces"):
        assert set(fuels["Fuel"]) == set(gens["Fuel"])
        fuel_emission_map = fuels.copy()
        fuel_emission_map = fuel_emission_map.set_index("Fuel")

        gens["co2_emission_rate"] = gens["Heat_Rate_MMBTU_per_MWh"] * gens["Fuel"].map(
            fuel_emission_map["CO2_content_tons_per_MMBtu"]
        )

        # Make the partial CES credit equal to 1 ton minus the emissions rate, but
        # don't include coal plants

        partial_ces = 1 - gens["co2_emission_rate"]

        gens.loc[
            ~(gens["Resource"].str.contains("coal"))
            & (gens["STOR"] == 0)
            & (gens["FLEX"] == 0),
            # & ~(gens["Resource"].str.contains("battery"))
            # & ~(gens["Resource"].str.contains("load_shifting")),
            "CES",
        ] = partial_ces.round(3)
    # else:
    #     gen_clusters = add_genx_model_tags(gen_clusters, settings)

    gens["Zone"] = gens["Zone"]
    gens["Cap_Size"] = gens["Cap_size"]
    gens["Fixed_OM_Cost_per_MWyr"] = gens["Fixed_OM_Cost_per_MWyr"]
    gens["Fixed_OM_Cost_per_MWhyr"] = gens["Fixed_OM_Cost_per_MWhyr"]
    gens["Inv_Cost_per_MWyr"] = gens["Inv_Cost_per_MWyr"]
    gens["Inv_Cost_per_MWhyr"] = gens["Inv_Cost_per_MWhyr"]
    gens["Var_OM_Cost_per_MWh"] = gens["Var_OM_Cost_per_MWh"]
    # gens["Var_OM_Cost_per_MWh_In"] = gens["Var_OM_Cost_per_MWh_in"]
    gens["Start_Cost_per_MW"] = gens["Start_Cost_per_MW"]
    gens["Start_Fuel_MMBTU_per_MW"] = gens["Start_fuel_MMBTU_per_MW"]
    gens["Heat_Rate_MMBTU_per_MWh"] = gens["Heat_Rate_MMBTU_per_MWh"]
    gens["Min_Power"] = gens["Min_Power"]
    # gens["Self_Disch"] = gens["Self_disch"]
    # gens["Eff_Up"] = gens["Eff_up"]
    # gens["Eff_Down"] = gens["Eff_down"]
    # gens["Ramp_Up_Percentage"] = gens["Ramp_Up_percentage"]
    # gens["Ramp_Dn_Percentage"] = gens["Ramp_Dn_percentage"]
    # gens["Up_Time"] = gens["Up_time"]
    # gens["Down_Time"] = gens["Down_time"]
    # gens["Max_Flexible_Demand_Delay"] = gens["Max_DSM_delay"]
    gens["technology"] = gens["Resource"]
    gens["Resource"] = (
        gens["region"] + "_" + gens["technology"] + "_" + gens["cluster"].astype(str)
    )

    return gens


def check_min_power_against_variability(gen_clusters, resource_profile):
    min_gen_levels = resource_profile.min()

    assert len(min_gen_levels) == len(
        gen_clusters
    ), "The number of hourly resource profiles does not match the number of resources"
    # assert (
    #     gen_clusters["Min_Power"].isna().any() is False
    # ), (
    #     "At least one Min_Power value in 'gen_clusters' is null before checking against"
    #     " resource variability"
    # )

    min_gen_levels.index = gen_clusters.index

    gen_clusters["Min_Power"] = gen_clusters["Min_Power"].combine(min_gen_levels, min)

    # assert (
    #     gen_clusters["Min_Power"].isna().any() is False
    # ), (
    #     "At least one Min_Power value in 'gen_clusters' is null. Values were fine "
    #     "before checking against resoruce variability"
    # )

    return gen_clusters


def set_must_run_generation(
    gen_variability: pd.DataFrame, must_run_techs: List[str] = None
) -> pd.DataFrame:
    """Set the generation of must run resources -- those tagged as "MUST_RUN" in
     generators data -- to 1 in all hours.

    Parameters
    ----------
    gen_variability : pd.DataFrame
        Columns are names of resources, rows are floats from 0-1
    must_run_techs : List[str], optional
        Names of the resources selected as must run, by default None

    Returns
    -------
    pd.DataFrame
        Modified version of gen_variability with any must run techs having a value of 1
        in all rows.

    Examples
    --------
    In-memory dataframe:

    >>> gen_variability = pd.DataFrame({
        "gen_1": [0.5, 0.6, 0.7],
        "gen_2": [0.8, 0.9, 1.0],
        "gen_3": [0.0, 0.0, 0.0]
    })
    >>> must_run_techs = ["gen_3"]
    >>> expected_output = pd.DataFrame({
        "gen_1": [0.5, 0.6, 0.7],
        "gen_2": [0.8, 0.9, 1.0],
        "gen_3": [1.0, 1.0, 1.0]
    })
    >>> assert set_must_run_generation(gen_variability, must_run_techs).equals(expected_output)
    """
    for tech in must_run_techs or []:
        if tech not in gen_variability.columns:
            logger.warning(
                f"Trying to set {tech} as a must run resource (hourly generation is) "
                "always 1), but it was not found in the generation variability dataframe."
            )
            continue
        gen_variability.loc[:, tech] = 1.0

    return gen_variability


def calc_emissions_ces_level(network_df, load_df, settings):
    # load_cols = [col for col in load_df.columns if "Load" in col]
    total_load = load_df.sum().sum()

    emissions_limit = settings["emissions_ces_limit"]

    if emissions_limit is not None:
        try:
            ces_value = 1 - (emissions_limit / total_load)
        except TypeError:
            print(emissions_limit, total_load)
            raise TypeError
        network_df["CES"] = ces_value
        network_df["CES"] = network_df["CES"].round(3)

        return network_df
    else:
        return network_df


def fix_min_power_values(
    resource_df: pd.DataFrame,
    gen_profile_df: pd.DataFrame,
    min_power_col: str = "Min_Power",
) -> pd.DataFrame:
    """Fix potentially erroneous min power values for resources with variable generation
    profiles. Any min power values that are higher than the lowest hourly generation
    will be adjusted down to match the lowest hourly generation.


    Parameters
    ----------
    resource_df : pd.DataFrame
        Records of generators/resources. Row order should match column order of
        `gen_profile_df`.
    gen_profile_df : pd.DataFrame
        Hourly generation values for all generators/resources. Column order should match
        row order in `resource_df`.
    min_power_col : str
        Column in `resource_df` that stores the minimum generation power of each
        resource. Default value is "Min_Power".

    Returns
    -------
    pd.DataFrame
        A modified version of `resource_df`. Any rows with minimum power larger than
        hourly generation are adjusted down to match the smallest hourly generation.
    """
    if min_power_col not in resource_df.columns:
        raise ValueError(
            f"When variable generation values against resource min power, the column "
            f"{min_power_col} was not found in the resource dataframe."
        )

    if resource_df.shape[0] != gen_profile_df.shape[1]:
        raise ValueError(
            "When trying to fix min power values, the number of resource dataframe rows"
            f" ({resource_df.shape[0]} rows) does not match the number of variable "
            f"profiles columns ({gen_profile_df.shape[1]} columns)."
        )

    resource_df = resource_df.reset_index(drop=True)
    resource_df.loc[:, "unadjusted_min_power"] = resource_df.loc[:, min_power_col]
    _gen_profile = gen_profile_df.copy(deep=True)
    _gen_profile
    gen_profile_min = _gen_profile.min().reset_index(drop=True)
    mask = (resource_df[min_power_col].fillna(0) > gen_profile_min).values

    logger.debug(
        f"{sum(mask)} resources have {min_power_col} larger than hourly generation."
    )

    resource_df.loc[mask, min_power_col] = gen_profile_min[mask].round(3)

    return resource_df


def min_cap_req(settings: dict) -> pd.DataFrame:
    """Create a dataframe of minimum capacity requirements for GenX

    Parameters
    ----------
    settings : dict
        Dictionary with user settings. Should include the key `MinCapReg` with nested
        keys of `MinCapTag_*`, then further nested keys `description` and `min_mw`. The
        `MinCapTag_*` should also be listed as values under `model_tag_names`. Any
        technologies eligible for each of the `MinCapTag_*` should have `model_tag_values`
        of 1.

    Returns
    -------
    pd.DataFrame
        A dataframe with minimum capacity constraints formatted for GenX. If `MinCapReq`
        is not included in the settings dictionary it will return None.

    Raises
    ------
    KeyError
        If a `MinCapTag_*` is included under `MinCapReq` but not included in `model_tag_names`
        the function will raise an error.
    """

    c_num = []
    description = []
    min_mw = []

    # if settings.get("MinCapReq"):
    for cap_tag, values in (settings.get("MinCapReq", {}) or {}).items():
        if cap_tag not in settings.get("model_tag_names", []):
            raise KeyError(
                f"The minimum capacity tag {cap_tag} is listed in the settings "
                "'MinCapReq' but not under 'model_tag_names'. You must add it to "
                "'model_tag_names' for the column to appear in Generators_data.csv."
            )

        # It's easy to forget to add all the necessary column names to the
        # generators_columns list in settings.
        if cap_tag not in settings.get("generator_columns", []) and isinstance(
            settings.get("generator_columns"), list
        ):
            settings["generator_columns"].append(cap_tag)

        c_num.append(cap_tag.split("_")[1])
        description.append(values.get("description"))
        min_mw.append(values.get("min_mw"))

    min_cap_df = pd.DataFrame()
    min_cap_df["MinCapReqConstraint"] = c_num
    min_cap_df["Constraint_Description"] = description
    min_cap_df["Min_MW"] = min_mw

    if not min_cap_df.empty:
        return min_cap_df
    else:
        return None


def max_cap_req(settings: dict) -> pd.DataFrame:
    """Create a dataframe of maximum capacity requirements for GenX

    Parameters
    ----------
    settings : dict
        Dictionary with user settings. Should include the key `MinCapReg` with nested
        keys of `MaxCapTag_*`, then further nested keys `description` and `min_mw`. The
        `MaxCapTag_*` should also be listed as values under `model_tag_names`. Any
        technologies eligible for each of the `MaxCapTag_*` should have `model_tag_values`
        of 1.

    Returns
    -------
    pd.DataFrame
        A dataframe with maximum capacity constraints formatted for GenX. If `MaxCapReq`
        is not included in the settings dictionary it will return None.

    Raises
    ------
    KeyError
        If a `MaxCapTag_*` is included under `MaxCapReq` but not included in `model_tag_names`
        the function will raise an error.
    """

    c_num = []
    description = []
    max_mw = []

    # if settings.get("MaxCapReq"):
    for cap_tag, values in (settings.get("MaxCapReq", {}) or {}).items():
        if cap_tag not in settings.get("model_tag_names", []):
            raise KeyError(
                f"The maximum capacity tag {cap_tag} is listed in the settings "
                "'MaxCapReq' but not under 'model_tag_names'. You must add it to "
                "'model_tag_names' for the column to appear in Generators_data.csv."
            )

        # It's easy to forget to add all the necessary column names to the
        # generators_columns list in settings.
        if cap_tag not in settings.get("generator_columns", []) and isinstance(
            settings.get("generator_columns"), list
        ):
            settings["generator_columns"].append(cap_tag)

        c_num.append(cap_tag.split("_")[1])
        description.append(values.get("description"))
        max_mw.append(values.get("max_mw"))

    max_cap_df = pd.DataFrame()
    max_cap_df["MaxCapReqConstraint"] = c_num
    max_cap_df["Constraint_Description"] = description
    max_cap_df["Max_MW"] = max_mw

    if not max_cap_df.empty:
        return max_cap_df
    else:
        return None


def check_resource_tags(df: pd.DataFrame) -> pd.DataFrame:
    """Check complete generators dataframe to make sure each resource is assigned one,
    and only one, resource tag.

    Parameters
    ----------
    df : pd.DataFrame
        Resource clusters. Should have columns "technology" and "region" in addition
        to the resource tag columns expected by GenX.

    Returns
    -------
    pd.DataFrame
        An unaltered version of the input dataframe.
    """
    tags = [t for t in RESOURCE_TAGS if t in df.columns]
    df_copy = df.loc[:, ["technology", "region"] + tags].copy()
    df_copy[tags] = df_copy[tags].where(df_copy[tags] == 0, 1)
    if not (df_copy[tags].sum(axis=1) == 1).all():
        for idx, row in df_copy.iterrows():
            num_tags = row[tags].sum()
            if num_tags == 0:
                logger.warning(
                    "\n*************************\n"
                    f"The resource {row['technology']} in region {row['region']} does "
                    "not have any assigned resource tags. Check the 'model_tag_values' and "
                    "'regional_tag_values' parameters in your settings file to make sure"
                    "it is assigned one resource tag type from this list:\n\n"
                    f"{RESOURCE_TAGS}\n"
                )
            if num_tags > 1:
                s = row[tags]
                _tags = list(s[s == 1].index)
                logger.warning(
                    "\n*************************\n"
                    f"The resource {row['technology']} in region {row['region']} is "
                    f"assigned {num_tags} resource tags ({_tags}). Check the 'model_tag_values'"
                    " and 'regional_tag_values' parameters in your settings file to make"
                    " sure it is assigned only one resource tag type from this list:\n\n"
                    f"{RESOURCE_TAGS}\n"
                )

        raise ValueError(
            "Use the warnings above to fix the resource tags in your settings file."
        )
    return df


def hydro_energy_to_power(
    df: pd.DataFrame,
    default_factor: float = None,
    regional_factors: Dict[str, float] = {},
) -> pd.DataFrame:
    """Calculate the hydro energy to power ratio. Uses average hydro inflow rate and
    multiplied by a factor to calculate the rated number of hours of reservoir hydro
    storage at peak discharge power output. Value minimum is 1.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame of resources. Hydro resources should be identified with a "HYDRO" tag
        value of 1
    default_factor : float, optional
        Hydro factor used to scale average inflow rate, by default None
    regional_factors : Dict[str, float], optional
        Specific regional hydro factors used to scale average inflow rate, by default {}

    Returns
    -------
    pd.DataFrame
        Copy of the input dataframe with the new column "Hydro_Energy_to_Power_Ratio"
    """
    if "HYDRO" not in df.columns:
        logger.warning(
            "Generators do not have a column 'HYDRO', so no hydro energy to power ratio is calculated."
        )
        return df
    if not default_factor and not regional_factors:
        logger.warning(
            "No hydro factors have been included in the settings, so no hydro energy to power ratio is calculated."
        )
        return df
    hydro_mask = df["HYDRO"] == 1
    avg_inflow = (
        make_generator_variability(df).mean().reset_index(drop=True) * default_factor
    ).loc[hydro_mask]
    df.loc[hydro_mask, "Hydro_Energy_to_Power_Ratio"] = avg_inflow.where(
        avg_inflow > 1, 1
    )

    for region, factor in (regional_factors or {}).items():
        region_mask = df["region"] == region
        if region_mask.any():
            avg_inflow = (
                make_generator_variability(df).mean().reset_index(drop=True) * factor
            ).loc[hydro_mask & region_mask]
            df.loc[(df["HYDRO"] == 1) & region_mask, "Hydro_Energy_to_Power_Ratio"] = (
                avg_inflow.where(avg_inflow > 1, 1)
            )
    df["Hydro_Energy_to_Power_Ratio"] = df["Hydro_Energy_to_Power_Ratio"].fillna(0)
    return df


def rename_gen_cols(
    df: pd.DataFrame, rename_cols: Dict[str, str] = None
) -> pd.DataFrame:
    """Rename columns in the generators data file.

    By default the only rename so far is "existing_mwh" to "Existing_Cap_MWh".

    Parameters
    ----------
    df : pd.DataFrame
        Final dataframe of generating resources.
    rename_cols : Dict[str, str], optional
        Additional rename pairs, by default None

    Returns
    -------
    pd.DataFrame
        Identical to input dataframe except for renamed columns.
    """

    rename = {
        "capacity_mwh": "Existing_Cap_MWh",
    }
    if rename_cols:
        rename.update(rename_cols)

    df = df.rename(columns=rename, errors="ignore")

    return df


def add_co2_costs_to_o_m(df: pd.DataFrame) -> pd.DataFrame:
    """Add CO2 pipeline annuity/FOM to fixed O&M and cost per MWh to variable O&M
    using the column names for GenX

    Parameters
    ----------
    df : pd.DataFrame
        Generators dataframe. May include columns "co2_cost_mwh", "co2_pipeline_annuity_mw",
        and "co2_o_m_mw". Must include columns "Var_OM_Cost_per_MWh",
        "Inv_Cost_per_MWyr", and "Fixed_OM_Cost_per_MWyr"

    Returns
    -------
    pd.DataFrame
        Modified version of original df with CO2 pipeline annuity/O&M added to plant
        costs
    """
    if "co2_cost_mwh" in df.columns:
        df["Var_OM_Cost_per_MWh"] += df["co2_cost_mwh"].fillna(0)
    if "co2_pipeline_annuity_mw" in df.columns:
        df["Inv_Cost_per_MWyr"] += df["co2_pipeline_annuity_mw"].fillna(0)
    if "co2_o_m_mw" in df.columns:
        df["Fixed_OM_Cost_per_MWyr"] += df["co2_o_m_mw"].fillna(0)
    if "co2_pipeline_capex_mw" in df.columns:
        df["capex_mw"] += df["co2_pipeline_capex_mw"].fillna(0)

    return df


def cap_retire_within_period(
    gens: pd.DataFrame, first_year: int, last_year: int, capacity_col: str
) -> pd.Series:
    retired_cap = (
        gens.query("retirement_year <= @last_year and retirement_year >= @first_year")
        .groupby("Resource", as_index=False)[[capacity_col, "capacity_mwh"]]
        .sum()
    ).rename(
        columns={
            capacity_col: "Min_Retired_Cap_MW",
            "capacity_mwh": "Min_Retired_Energy_Cap_MW",
        }
    )
    retired_cap["Min_Retired_Charge_Cap_MW"] = 0

    return retired_cap


def check_vre_profiles(
    gen_df: pd.DataFrame,
    gen_var_df: pd.DataFrame,
    vre_cols: List[str] = ["VRE", "HYDRO"],
):
    """Check generation profiles of VRE resources to ensure they are variable. Alert the
    user with a logger warning if any are not.

    Parameters
    ----------
    gen_df : pd.DataFrame
        Dataframe of generators. Must have column "Resource".
    gen_var_df : pd.DataFrame
        Dataframe of hourly generation for each resource. Column names are from the
        "Resource" column of `gen_df`.
    vre_cols : List[str]
        List of boolean columns indicating inclusion in the group of variable resources.
        By default, ["VRE", "HYDRO"].
    """
    var_gen_names = []
    vre_cols = [c for c in vre_cols if c in gen_df.columns]
    if vre_cols:
        for c in vre_cols:
            var_gen_names.extend(gen_df.loc[gen_df[c] == 1, "Resource"].to_list())

        vre_std = gen_var_df[var_gen_names].std()
        if (vre_std == 0).any():
            non_variable = vre_std.loc[vre_std == 0].index.to_list()
            logger.warning(
                f"The variable resources {non_variable} have non-variable generation profiles."
            )


def update_newbuild_canretire(df: pd.DataFrame) -> pd.DataFrame:
    """Update the New_Build and Can_Retire columns in generator data.

    If Can_Retire column doesn't exist, creates it based on New_Build values:
    - Can_Retire = 1 if New_Build != -1, else 0
    - New_Build = 1 if New_Build == 1, else 0

    Parameters
    ----------
    df : pd.DataFrame
        Generator data with New_Build column

    Returns
    -------
    pd.DataFrame
        Modified dataframe with updated Can_Retire and New_Build columns
    """
    if "New_Build" not in df.columns:
        logger.warning("New_Build column not found in generator data")
        return df

    if not df["New_Build"].isin([0, 1, -1]).all():
        logger.warning("New_Build column contains values other than 0, 1, or -1")
        return df

    logger.warning("Upgrading the Can_Retire and New_Build interface")
    df["Can_Retire"] = (df["New_Build"] != -1).astype(int)
    df["New_Build"] = (df["New_Build"] == 1).astype(int)

    return df


def filter_empty_columns(df: pd.DataFrame) -> List[str]:
    """Find columns that have at least one valid value.

    A column is considered valid if it has at least one value that is:
    - Non-zero
    - Not None (Python None)
    - Not "No_fuel" (string)

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe to check

    Returns
    -------
    List[str]
        List of column names that have at least one non-zero and non-None/"No_fuel" value
    """
    # Drop columns that are known to have lists or array values
    drop_cols = ["plant_id_eia", "unit_id_eia", "unit_id_pg", "generator_id", "profile"]
    df = df.drop(columns=drop_cols, errors="ignore")

    # Check for non-None values
    notnull_mask = df.notna().sum() > 0

    # Check for non-"No_fuel" string values
    string_notnone_mask = (
        df.applymap(lambda x: str(x) != "No_fuel" if pd.notna(x) else False).sum() > 0
    )

    # Check for non-zero values
    nonzero_mask = df.astype(bool).sum() > 0

    # Combine all masks
    valid_cols = nonzero_mask & notnull_mask & string_notnone_mask

    return list(valid_cols[valid_cols].index)


def create_resource_df(df: pd.DataFrame, resource_tag: str) -> pd.DataFrame:
    """Create a dataframe with resource-specific columns for a specific resource type.

    Filters generator data for a specific resource type and removes columns
    specific to other resource types. For storage and thermal resources,
    renames the resource tag column to 'Model'.

    Parameters
    ----------
    df : pd.DataFrame
        Generator data with resource type columns
    resource_tag : str
        Resource type tag (e.g., 'STOR', 'THERM', etc.)

    Returns
    -------
    pd.DataFrame
        Filtered and restructured data for the specific resource type.
        Returns empty DataFrame if resource_tag not in columns.

    """
    logger.info(f"Creating resource data file for {resource_tag}")

    if resource_tag not in df.columns:
        logger.warning(f"Resource tag {resource_tag} not found in dataframe columns")
        return pd.DataFrame()

    # Filter rows for this resource type
    resource_df = df[df[resource_tag] > 0].copy()

    # Get columns to keep for this resource type
    resource_specific_cols = set(RESOURCE_COLUMNS[resource_tag])

    # Find columns with non-zero values
    valid_cols = (
        set(filter_empty_columns(resource_df))
        | set(DEFAULT_COLS)
        | set(resource_specific_cols)
    )
    logger.debug(f"Found {len(valid_cols)} valid columns for {resource_tag}")

    # Get all columns specific to other resource types
    # i.e. columns that do not belong to this resource type
    other_resource_cols = {
        col
        for tag, cols in RESOURCE_COLUMNS.items()
        if tag != resource_tag
        for col in cols
    }
    # Add the resource tag column to the set of other resource columns
    other_resource_cols.add(resource_tag)

    # Find columns to remove: columns that are in other_resource_cols
    # but NOT in resource_specific_cols
    cols_to_remove = (other_resource_cols - resource_specific_cols) & set(df.columns)

    # List of columns to keep
    cols_to_keep = [
        col for col in df.columns if col in valid_cols and col not in cols_to_remove
    ]

    # In the case of STOR and THERM, the resource tag column is also used as
    # 'model' type (e.g., SYM or ASYM for storage, and COMMIT or NOCOMMIT for thermal)
    # Therefore, we rename the resource tag column to 'Model'
    if resource_tag in {"STOR", "THERM"}:
        logger.debug(f"Renaming {resource_tag} column to 'Model'")
        resource_df = resource_df.rename(columns={resource_tag: "Model"})
        cols_to_keep.append("Model")

    # Final check that ALL the resource specific columns are present
    missing_cols = resource_specific_cols - set(cols_to_keep)
    if missing_cols:
        logger.warning(f"Missing columns for {resource_tag}: {missing_cols}")

    # Make sure the first columns are DEFAULT_COLS defined at the top of this file
    remaining_cols = [col for col in cols_to_keep if col not in DEFAULT_COLS]
    final_cols = [col for col in DEFAULT_COLS if col in cols_to_keep] + remaining_cols

    # Return the filtered and restructured dataframe
    return resource_df[final_cols]


def create_policy_df(df: pd.DataFrame, policy_info: Dict[str, str]) -> pd.DataFrame:
    """Create a dataframe with policy-specific columns for a specific policy type.

    Filters generator data for a specific policy type and removes columns
    specific to other policy types.

    Parameters
    ----------
    df : pd.DataFrame
        Generator data with policy type columns
    policy_info : Dict[str, str]
        Policy type configuration

    Returns
    -------
    pd.DataFrame
        Dataframe containing the policy tag columns
    """
    logger.info(f"Creating policy data file for {policy_info['newtag']}")

    if df.empty:
        logger.warning(f"DataFrame is empty, skipping policy {policy_info['newtag']}")
        return pd.DataFrame()

    # Check if the Resource column exists
    if "Resource" not in df.columns:
        raise KeyError("The 'Resource' column is missing from the dataframe.")

    # Check that ESR_, MinCapTag_, and MaxCapTag_ columns only have values equal to 0 or 1
    # and CapRes_ columns only have values between 0 and 1
    for col in df.columns:
        if (
            col.startswith("ESR_")
            or col.startswith("MinCapTag_")
            or col.startswith("MaxCapTag_")
        ):
            if not df[col].isin([0, 1]).all():
                raise ValueError(f"The column {col} can only have values 0 or 1.")
        elif col.startswith("CapRes_"):
            if not (df[col] >= 0).all() or not (df[col] <= 1).all():
                raise ValueError(
                    f"The column {col} can only have values between 0 and 1."
                )

    # Check if any columns start with the policy tag
    if not any(col.startswith(policy_info["oldtag"]) for col in df.columns):
        logger.debug(
            f"No columns start with {policy_info['oldtag']}, skipping policy {policy_info['newtag']}"
        )
        return pd.DataFrame()

    # Slice dataframe to include only Resource column and columns starting with oldtag
    policy_cols = [
        col
        for col in df.columns
        if col == "Resource" or col.startswith(policy_info["oldtag"])
    ]
    policy_df = df[policy_cols].copy()

    # Keep only rows with at least one non-zero value in policy columns
    policy_data_cols = policy_df.select_dtypes(
        include=["number"]
    ).columns  # All numeric columns
    policy_df = policy_df[policy_df[policy_data_cols].gt(0).any(axis=1)]

    # Rename columns replacing oldtag with newtag
    rename_dict = {
        col: col.replace(policy_info["oldtag"], policy_info["newtag"])
        for col in policy_df.columns
        if col.startswith(policy_info["oldtag"])
    }
    policy_df.rename(columns=rename_dict, inplace=True)

    # Remove policy columns from original dataframe
    df.drop(
        columns=[col for col in policy_data_cols if col in df.columns], inplace=True
    )

    return policy_df


def create_multistage_df(df: pd.DataFrame, multistage_cols: List[str]) -> pd.DataFrame:
    """Create multistage data file from generator data.

    Parameters
    ----------
    df : pd.DataFrame
        Generator data
    multistage_cols : List[str]
        List of column names for multistage data

    Returns
    -------
    pd.DataFrame
        DataFrame containing Resource column and multistage data
    """
    logger.info("Creating multistage data file")

    # Select Resource column and multistage columns
    cols_to_keep = ["Resource"] + [col for col in multistage_cols if col in df.columns]
    multistage_df = df[cols_to_keep].copy()

    # Remove multistage columns from original dataframe
    df.drop(columns=[col for col in multistage_cols if col in df.columns], inplace=True)

    return multistage_df


def split_generators_data(
    folder_structure: FolderStructure, gen_data: pd.DataFrame
) -> List[GenXInputData]:
    """Split generator data DataFrame into resource-specific files.

    Parameters
    ----------
    gen_data : pd.DataFrame
        DataFrame containing generator data

    Returns
    -------
    List[GenXResourceData]
        List of GenXResourceData objects containing the resource-specific data
    """

    # Paths to the folders to write the data to
    resource_folder = folder_structure.resource
    policy_assignments_folder = folder_structure.policy_assignments

    # Create a copy of the generator data
    generators_df = gen_data.copy()

    # Drop the R_ID column if it exists, GenX will assign IDs internally
    if "R_ID" in generators_df.columns:
        generators_df = generators_df.drop(columns=["R_ID"])

    # Check if New_Build and Can_Retire columns exist
    if (
        "New_Build" in generators_df.columns
        and "Can_Retire" not in generators_df.columns
    ):
        generators_df = update_newbuild_canretire(generators_df)

    # Process each POLICY_TAGS (defined at the top of this file) and return a dataframe
    # with the policy files to be written out
    resource_data = []
    for policy_tag, policy_info in POLICY_TAGS:
        policy_df = create_policy_df(generators_df, policy_info)
        if not policy_df.empty:
            out_file = POLICY_TAGS_FILENAMES[policy_tag]
            resource_data.append(
                GenXInputData(
                    tag=f"RES_{policy_tag}",
                    folder=policy_assignments_folder,
                    file_name=out_file,
                    dataframe=policy_df,
                )
            )

    # Process multistage data
    multistage_df = create_multistage_df(generators_df, MULTISTAGE_COLS)
    if not multistage_df.empty:
        out_file = "Resource_multistage_data.csv"
        resource_data.append(
            GenXInputData(
                tag="MULTISTAGE",
                folder=resource_folder,
                file_name=out_file,
                dataframe=multistage_df,
            )
        )

    # Process each RESOURCE_TAGS defined at the top of this file
    for resource_tag in RESOURCE_TAGS:
        out_file = RESOURCE_FILENAMES[resource_tag]
        resource_df = create_resource_df(generators_df, resource_tag)
        if not resource_df.empty:
            resource_data.append(
                GenXInputData(
                    tag=resource_tag,
                    folder=resource_folder,
                    file_name=out_file,
                    dataframe=resource_df,
                )
            )
        else:
            logger.info(f"No data found for resource tag {resource_tag}")

    # Return a list of GenXResourceData objects
    return resource_data


def process_genx_data(
    case_folder: Path, genx_data_dict: Dict[str, pd.DataFrame]
) -> List[GenXInputData]:
    """Process genx data dictionary and return a list of GenXResourceData objects.

    Parameters
    ----------
    case_folder : Path
        Path to the case folder
    genx_data_dict : Dict[str, pd.DataFrame]
        Dictionary of resource-specific data

    Returns
    -------
    List[GenXResourceData]
        List of GenXResourceData objects to be written out
    """

    resource_folder = case_folder / "resources"
    policy_assignments_folder = resource_folder / "policy_assignments"
    system_folder = case_folder / "system"
    policy_folder = case_folder / "policies"

    # Create folder structure
    folders = FolderStructure(
        resource=resource_folder,
        policy_assignments=policy_assignments_folder,
        system=system_folder,
        policy=policy_folder,
    )

    # List of GenXInputData objects to be written out
    genx_data = []

    # Process generator data separately
    genx_data.extend(split_generators_data(folders, genx_data_dict["gen_data"]))

    # Add all other data to the list
    genx_data.extend(
        [
            # System folder data
            GenXInputData(
                tag="GENERATORS_VARIABILITY",
                folder=system_folder,
                file_name="Generators_variability.csv",
                dataframe=genx_data_dict.get("gen_variability", pd.DataFrame()),
            ),
            GenXInputData(
                tag="DEMAND",
                folder=system_folder,
                file_name="Demand_data.csv",
                dataframe=genx_data_dict.get("demand_data", pd.DataFrame()),
            ),
            GenXInputData(
                tag="PERIOD_MAP",
                folder=system_folder,
                file_name="Period_map.csv",
                dataframe=genx_data_dict.get("period_map", pd.DataFrame()),
            ),
            GenXInputData(
                tag="REP_PERIOD",
                folder=system_folder,
                file_name="Representative_Period.csv",
                dataframe=genx_data_dict.get("rep_period", pd.DataFrame()),
            ),
            GenXInputData(
                tag="NETWORK",
                folder=system_folder,
                file_name="Network.csv",
                dataframe=genx_data_dict.get("network", pd.DataFrame()),
            ),
            GenXInputData(
                tag="OP_RESERVES",
                folder=system_folder,
                file_name="Operational_reserves.csv",
                dataframe=genx_data_dict.get("op_reserves", pd.DataFrame()),
            ),
            GenXInputData(
                tag="FUELS",
                folder=system_folder,
                file_name="Fuels_data.csv",
                dataframe=genx_data_dict.get("fuels", pd.DataFrame()),
            ),
            # Policy folder data
            GenXInputData(
                tag="ESR",
                folder=policy_folder,
                file_name="Energy_share_requirement.csv",
                dataframe=genx_data_dict.get("esr", pd.DataFrame()),
            ),
            GenXInputData(
                tag="CAP_RESERVES",
                folder=policy_folder,
                file_name="Capacity_reserve_margin.csv",
                dataframe=genx_data_dict.get("cap_reserves", pd.DataFrame()),
            ),
            GenXInputData(
                tag="CO2_CAP",
                folder=policy_folder,
                file_name="CO2_cap.csv",
                dataframe=genx_data_dict.get("co2_cap", pd.DataFrame()),
            ),
            GenXInputData(
                tag="MIN_CAP",
                folder=policy_folder,
                file_name="Minimum_capacity_requirement.csv",
                dataframe=genx_data_dict.get("min_cap", pd.DataFrame()),
            ),
            GenXInputData(
                tag="MAX_CAP",
                folder=policy_folder,
                file_name="Maximum_capacity_requirement.csv",
                dataframe=genx_data_dict.get("max_cap", pd.DataFrame()),
            ),
        ]
    )

    return genx_data


def process_genx_data_old_format(
    case_folder: Path, genx_data_dict: Dict[str, pd.DataFrame]
) -> List[GenXInputData]:
    """Process genx data dictionary and return a list of GenXResourceData objects. Use
    for GenX versions prior to 0.4.0.

    Parameters
    ----------
    case_folder : Path
        Path to the case folder
    genx_data_dict : Dict[str, pd.DataFrame]
        Dictionary of resource-specific data

    Returns
    -------
    List[GenXResourceData]
        List of GenXResourceData objects to be written out
    """

    # Add 'None' column to fuels data if it doesn't exist -- causes error in old GenX otherwise
    if (
        "fuels" in genx_data_dict
        and "None" not in genx_data_dict.get("fuels", pd.DataFrame()).columns
    ):
        genx_data_dict["fuels"]["None"] = 0

    if "demand_data" in genx_data_dict:
        genx_data_dict["demand_data"].columns = genx_data_dict[
            "demand_data"
        ].columns.str.replace("Demand_", "Load_")

    # List of GenXInputData objects to be written out
    genx_data = []

    # Add all other data to the list
    genx_data.extend(
        [
            GenXInputData(
                tag="GENERATORS_DATA",
                folder=case_folder,
                file_name="Generators_data.csv",
                dataframe=genx_data_dict.get("gen_data", pd.DataFrame()),
            ),
            GenXInputData(
                tag="GENERATORS_VARIABILITY",
                folder=case_folder,
                file_name="Generators_variability.csv",
                dataframe=genx_data_dict.get("gen_variability", pd.DataFrame()),
            ),
            GenXInputData(
                tag="DEMAND",
                folder=case_folder,
                file_name="Load_data.csv",
                dataframe=genx_data_dict.get("demand_data", pd.DataFrame()),
            ),
            GenXInputData(
                tag="PERIOD_MAP",
                folder=case_folder,
                file_name="Period_map.csv",
                dataframe=genx_data_dict.get("period_map", pd.DataFrame()),
            ),
            GenXInputData(
                tag="REP_PERIOD",
                folder=case_folder,
                file_name="Representative_Period.csv",
                dataframe=genx_data_dict.get("rep_period", pd.DataFrame()),
            ),
            GenXInputData(
                tag="NETWORK",
                folder=case_folder,
                file_name="Network.csv",
                dataframe=genx_data_dict.get("network", pd.DataFrame()),
            ),
            GenXInputData(
                tag="OP_RESERVES",
                folder=case_folder,
                file_name="Operational_reserves.csv",
                dataframe=genx_data_dict.get("op_reserves", pd.DataFrame()),
            ),
            GenXInputData(
                tag="FUELS",
                folder=case_folder,
                file_name="Fuels_data.csv",
                dataframe=genx_data_dict.get("fuels", pd.DataFrame()),
            ),
            GenXInputData(
                tag="ESR",
                folder=case_folder,
                file_name="Energy_share_requirement.csv",
                dataframe=genx_data_dict.get("esr", pd.DataFrame()),
            ),
            GenXInputData(
                tag="CAP_RESERVES",
                folder=case_folder,
                file_name="Capacity_reserve_margin.csv",
                dataframe=genx_data_dict.get("cap_reserves", pd.DataFrame()),
            ),
            GenXInputData(
                tag="CO2_CAP",
                folder=case_folder,
                file_name="CO2_cap.csv",
                dataframe=genx_data_dict.get("co2_cap", pd.DataFrame()),
            ),
            GenXInputData(
                tag="MIN_CAP",
                folder=case_folder,
                file_name="Minimum_capacity_requirement.csv",
                dataframe=genx_data_dict.get("min_cap", pd.DataFrame()),
            ),
            GenXInputData(
                tag="MAX_CAP",
                folder=case_folder,
                file_name="Maximum_capacity_requirement.csv",
                dataframe=genx_data_dict.get("max_cap", pd.DataFrame()),
            ),
        ]
    )

    return genx_data



"""
Flexible methods to cluster/aggregate renewable projects
"""

import logging
import operator
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import pyarrow.parquet as pq
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.preprocessing import StandardScaler

from powergenome.resource_clusters import MERGE
from powergenome.util import deep_freeze_args, snake_case_str

logger = logging.getLogger(__name__)


def load_site_profiles(path: Path, site_ids: List[str]) -> pd.DataFrame:
    suffix = path.suffix
    site_ids = [s.replace(".0", "") for s in site_ids]
    if suffix == ".parquet":
        df = pq.read_table(path, columns=site_ids).to_pandas()
    elif suffix == ".csv":
        df = pd.read_csv(path, usecols=site_ids)
    return df


def value_bin(
    s: pd.Series,
    bins: Union[int, List[float]] = None,
    q: Union[int, List[float]] = None,
    weights: pd.Series = None,
) -> pd.Series:
    """Create a series of bin labels to split feature values

    Feature data can be split into bins based directly on values using pandas.cut, or
    quantiles of the data using pandas.qcut. If both "q" and "weights" are passed then
    a weighted quantile will be calculated.

    Parameters
    ----------
    s : pd.Series
        Data values for a feature
    bins : Union[int, List[float]], optional
        Integer number of equal-width bins or the bin edges to use, by default None
    q : Union[int, List[float]], optional
        Integer number of quantile bins or the quantile bin edges to use, by default None
    weights : pd.Series, optional
        Optional weight of each feature data point, only used in conjunction with "q",
        by default None

    Returns
    -------
    pd.Series
        Labels for each data point in the feature
    """
    try:
        from statsmodels.stats.weightstats import DescrStatsW
    except ModuleNotFoundError as e:
        logger.error(
            "\n\nThe package 'statsmodels' must be installed for weighted quantile calculations. "
            "Install it using mamba, pip, or conda, then try again.\n\n"
        )
        raise ModuleNotFoundError(e)
    if s.empty:
        return []
    # if all values are very close pandas will give a binning error
    if np.allclose(s.mean(), s.max()):
        return np.ones_like(s)
    if q:
        if q == 1:
            return np.ones_like(s)
        if weights is None:
            labels = pd.qcut(s, q=q, duplicates="drop", labels=False)
            return labels
        elif (weights == 0).all():
            weights = pd.Series(np.ones_like(weights))

        if isinstance(q, int):  # Calc feature values of bin edges from quantiles
            q = np.linspace(0, 1, q + 1)
        q.sort()
        wq = DescrStatsW(data=s, weights=weights)
        bins = wq.quantile(probs=q, return_pandas=False)
        labels = pd.cut(s, bins=bins, duplicates="drop", include_lowest=True)

    elif bins:
        if isinstance(bins, list):
            bins.sort()
            if bins[0] > s.min():
                logger.warning(
                    f"The minimum bin value in one of your renewables_clusters is "
                    f"{bins[0]}, which is greater than the minimum value of the data series. "
                    f"The min bin edge is being set to {s.min()}"
                )
                bins[0] = s.min()

            if bins[-1] < s.max():
                logger.warning(
                    f"The maximum bin value in one of your renewables_clusters is "
                    f"{bins[-1]}, which is less than the maximum value of the data series. "
                    f"The max bin edge is being set to {s.max()}"
                )
                bins[-1] = s.max()

            labels = pd.cut(s, bins=bins, duplicates="drop", include_lowest=True)
        else:
            if bins == 1:
                return np.ones_like(s)
            else:
                labels = pd.cut(s, bins=bins, duplicates="drop", include_lowest=True)
    else:
        logger.warning(
            "One of your renewables_clusters uses the 'bin' option but doesn't include "
            "either the 'bins' or 'q' argument. One of these is required to bin sites "
            "by a numeric feature. No binning will be performed on this group of sites."
        )
        labels = np.ones_like(s)

    return labels


# @deep_freeze_args
# @lru_cache()
def agg_cluster_profile(s: pd.Series, n_clusters: int, **kwargs) -> np.ndarray:
    if len(s) == 0:
        return []
    if len(s) == 1:
        return np.array([0])
    if n_clusters <= 0:
        logger.warning(
            f"You have entered a n_clusters parameter that is less than or equal to 0 "
            "in the settings parameter renewables_clusters. n_clusters must be >= 1. "
            "Manually setting n_clusters to 1 in this case."
        )
        n_clusters = 1
    if n_clusters > len(s):
        logger.warning(
            f"You have entered a n_clusters parameter that is greater than the number of "
            "renewable sites for one grouping in the settings parameter renewables_clusters."
            "Manually setting n_clusters to the number of renewable sites in this case."
        )
        n_clusters = len(s)

    clust = AgglomerativeClustering(n_clusters=n_clusters).fit(
        np.array([x for x in np.array(s)])
    )
    labels = clust.labels_
    return labels


# @deep_freeze_args
# @lru_cache()
def agg_cluster_other(s: pd.Series, n_clusters: int, **kwargs) -> np.ndarray:
    if len(s) == 0:
        return []
    if len(s) == 1:
        return np.array([0])
    if n_clusters <= 0:
        logger.warning(
            f"You have entered a n_clusters parameter that is less than or equal to 0 "
            "in the settings parameter renewables_clusters. n_clusters must be >= 1. "
            "Manually setting n_clusters to 1 in this case."
        )
        n_clusters = 1
    if n_clusters > len(s):
        logger.warning(
            f"You have entered a n_clusters parameter that is greater than the number of "
            "renewable sites for one grouping in the settings parameter renewables_clusters."
            "Manually setting n_clusters to the number of renewable sites in this case."
        )
        n_clusters = len(s)
    clust = AgglomerativeClustering(n_clusters=n_clusters).fit(
        np.array(s).reshape(-1, 1)
    )
    labels = clust.labels_
    return labels


def kmeans_cluster_other(df: pd.DataFrame, n_clusters: int, **kwargs) -> np.ndarray:
    if len(df) == 0 or df.empty:
        return []
    if len(df) == 1:
        return np.array([0])
    if n_clusters <= 0:
        logger.warning(
            f"You have entered a n_clusters parameter that is less than or equal to 0 "
            "in the settings parameter renewables_clusters. n_clusters must be >= 1. "
            "Manually setting n_clusters to 1 in this case."
        )
        n_clusters = 1
    if n_clusters > len(df):
        logger.warning(
            f"You have entered a n_clusters parameter that is greater than the number of "
            "renewable sites for one grouping in the settings parameter renewables_clusters."
            "Manually assiging each site to its own cluster."
        )
        return np.arange(0, len(df))

    clust = KMeans(n_clusters=n_clusters, random_state=42, **kwargs).fit(
        StandardScaler().fit_transform(df)
    )
    labels = clust.labels_
    return labels


def cluster_sites_binned(
    data: pd.DataFrame,
    by: Union[str, List[str]],
    method: str,
    feature: Union[str, List[str]],
    n_clusters: Union[int, pd.Series, dict],
    **kwargs,
) -> pd.DataFrame:
    if data.empty:
        data["cluster"] = []
        return data

    if feature == "profile":
        if method not in ["agglomerative", "agg", "hierarchical"]:
            logger.warning(
                f"The renewables clustering method {method} cannot be used with the "
                "'profile' feature. Using agglomerative clustering instead."
            )
        func = agg_cluster_profile
    else:
        func = CLUSTER_METHOD_FUNCS[method]

    grouped = data.groupby(by)
    df_list = []
    first_label = 0
    if isinstance(n_clusters, int):
        _n_clusters = n_clusters
    for _, _df in grouped:
        if not _df.empty:
            if not isinstance(n_clusters, int):
                _n_clusters = n_clusters[_]
            labels = func(_df[feature], min(_n_clusters, len(_df)), **kwargs)
            labels += first_label
            first_label = max(labels) + 1
            _df["cluster"] = labels
            df_list.append(_df)
        else:
            _df = pd.DataFrame(columns=list(data.columns) + ["cluster"])
            df_list.append(_df)
    df = pd.concat(df_list)

    return df


def cluster_sites_no_bin(
    data: pd.DataFrame,
    method: str,
    feature: Union[str, List[str]],
    n_clusters: int,
    **kwargs,
) -> pd.DataFrame:
    if data.empty:
        data["cluster"] = []
        return data
    _data = data.copy()
    if feature == "profile":
        if method not in ["agglomerative", "agg", "hierarchical"]:
            logger.warning(
                f"The renewables clustering method {method} cannot be used with the "
                "'profile' feature. Using agglomerative clustering instead."
            )
        func = agg_cluster_profile
    else:
        func = CLUSTER_METHOD_FUNCS[method]

    if len(_data) == 1:
        labels = 1
        _data["cluster"] = labels
    else:
        labels = func(_data[feature], min(n_clusters, len(_data)), **kwargs)
        _data["cluster"] = labels

    return _data


CLUSTER_METHOD_FUNCS = {
    "agglomerative": agg_cluster_other,
    "agg": agg_cluster_other,
    "hierarchical": agg_cluster_other,
    "kmeans": kmeans_cluster_other,
}


def cluster_sites(binned: bool, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
    # kwargs.pop("method")
    if binned:
        return cluster_sites_binned(data, **kwargs)
    else:
        return cluster_sites_no_bin(data, **kwargs)


def value_filter(
    data: pd.DataFrame, feature: str, max_value: float = None, min_value: float = None
) -> pd.DataFrame:
    df = data.copy()
    if max_value:
        df = df.loc[df[feature] <= max_value, :]
    if min_value:
        df = df.loc[df[feature] >= min_value, :]

    return df


def min_capacity_mw(
    data: pd.DataFrame,
    min_cap: int = None,
    cap_col: str = "mw",
) -> pd.DataFrame:
    if "lcoe" not in data.columns:
        logger.warning(
            "You have selected a minimum capacity for one of the renewables_clusters "
            "groups. This feature is only available if the column 'lcoe' is included in "
            "your data file on renewable sites. No 'lcoe' column has been found, so all "
            "sites will be included in the final clusters."
        )
        return data
    if min_cap is None:
        return data

    _df = data.sort_values("lcoe")
    mask = np.ones(len(_df), dtype=bool)
    temp = (_df.loc[mask, cap_col].cumsum() < min_cap).values
    temp[temp.argmin()] = True
    mask[mask] = temp

    return _df.loc[mask, :]


def calc_cluster_values(
    df: pd.DataFrame,
    group: List[str] = None,
    sums: List[str] = MERGE["sums"],
    means: List[str] = MERGE["means"],
    weight: str = MERGE["weight"],
) -> pd.DataFrame:
    sums = [s for s in sums if s in df.columns]
    means = [m for m in means if m in df.columns]
    assert weight in df.columns
    df = df.reset_index(drop=True)
    df["weight"] = df[weight] / df[weight].sum()

    data = {}
    for s in sums:
        data[s] = df[s].sum()
    for m in means:
        data[m] = np.average(df[m], weights=df[weight])

    _df = pd.DataFrame(data, index=[0])
    profile = df.loc[0, "profile"] * df.loc[0, "weight"]
    for row in df.loc[1:, :].itertuples():
        profile += row.profile * row.weight

    profile /= df["weight"].sum()

    _df["profile"] = [profile]
    _df["cluster"] = df["cluster"].values[0]
    for g in group or []:
        _df["cluster"] = (
            str(_df["cluster"][0]) + f"_{g}_" + str(df[snake_case_str(g)].iloc[0])
        )

    return _df


CLUSTER_FUNCS = {
    "agglomerative": cluster_sites,
    "agg": cluster_sites,
    "hierarchical": cluster_sites,
}


def assign_site_cluster(
    renew_data: pd.DataFrame,
    profile_path: Path,
    regions: List[str],
    site_map: pd.DataFrame = None,
    min_capacity: int = None,
    filter: List[dict] = None,
    bin: List[dict] = None,
    group: List[str] = None,
    cluster: List[dict] = None,
    utc_offset: int = 0,
    **kwargs: Any,
) -> pd.DataFrame:
    """Use settings options to group individual renewable sites.
    Sites are located within a model region that may be a collection of base regions.
    Site metadata such as the cpa_id and any numeric or categorial features (e.g.
    interconnection cost or the name of a geographic subregion that the site is located
    in) are specified in `renew_data`. Based on the user settings, the full list of
    sites in a region is filtered (numeric), binned (numeric), grouped(categorical), and
    clustered (numeric).
    Parameters
    ----------
    renew_data : pd.DataFrame
        Data on each renewable site. Must contain columns "cpa_id", "region", and either
        "interconnect_annuity" or "interconnect_capex_mw". Other columns are numeric or
        categorical features. Example columns might be a geographic location such as county
        (categorical), or the capacity factor (numeric).
    profile_path : Path
        Path to a parquet or CSV file with generation profiles. Column names should
        either correspond to the "cpa_id" from `renew_data` or mapped IDs from `site_map`.
    regions : List[str]
        The regions from "region" column of `renew_data` that will be used.
    site_map : pd.DataFrame, optional
        An optional mapping of `cpa_id` to the column names in the generation profile file,
        by default None
    min_capacity : int, optional
        A minimum amount of capacity to include, by default None. Can only be used if
        the column "lcoe" is included in `renew_data`. Sites are sorted based on lcoe
        and the least cost sites that satisfy the minimum capacity are retained.
    filter : List[dict], optional
        A list of filter parameters, by default None. Each dictionary should have the keys
        "feature" and at least one of "min" and "max". Only sites in `renew_data` with
        values that satisfy the min/max parameters for the specified features are retained.
    bin : List[dict], optional
        A list of binning parameters, by default None. Each dictionary should have the keys
        "feature" and either "bins" or "q". "bins" and "q" are input parameters to the
        Pandas functions `cut` and `qcut` -- they can be integers or a list of bin edge
        values.
    group : List[str], optional
        A list of categorical features in `renew_data` to group the sites, by default None.
    cluster : List[dict], optional
        A list of clustering parameters, by default None. Each dictionary should have the
        keys "feature", "method", and "n_clusters". Supported methods are "agglomerative".
        The feature should either be a column from `renew_data` or "profile" (use the
        generation profiles as a feature).
    utc_offset : int, optional
        Hours offset from UTC, by default 0. Generation data will be shifted from UTC
        by this value.
    Returns
    -------
    pd.DataFrame
        The renewable sites included in the study with a column "cluster". Other columns
        for binning and grouping may also be included.
    Raises
    ------
    KeyError
        The key "feature" is not included in a "bin" set of parameters
    KeyError
        The feature is not included in `renew_data`
    TypeError
        The feature specified by a "bin" set of parameters is not numeric.
    Examples
    --------
    This is an example of the YAML settings.
    renewables_clusters:
    - region: ISNE
        technology: landbasedwind
        min_capacity: 10000
        filter:
          - feature: lcoe
            max: 100
        bin:
          - feature: interconnect_annuity
            weights: mw
            bins: 3 # This can be an integer or list of bin edges
            # q: [0, .25, .5, .75, 1.] # Integer number of quantiles (e.g. quartiles) or list of quantile edges
        group:
          - state
        cluster:
          - feature: profile
            method: agglomerative
            n_clusters: 2
    """
    data = renew_data.loc[renew_data["region"].isin(regions), :]
    group = [snake_case_str(g) for g in group or []]
    if group:
        data[group] = data[group].convert_dtypes()

    # Filter sites
    for filt in filter or []:
        data = value_filter(
            data=data,
            feature=snake_case_str(filt["feature"]),
            max_value=filt.get("max"),
            min_value=filt.get("min"),
        )
    if data.empty:
        data["cluster"] = []
        return data

    # Only keep X MW least cost capacity
    if min_capacity:
        data = min_capacity_mw(data, min_cap=min_capacity)
    if site_map is not None:
        # site_ids = [site_map.loc[i] for i in data["cpa_id"]]
        missing_site_ids = data.loc[
            data["cpa_id"].map(site_map).isna(), "cpa_id"
        ].to_list()
        if missing_site_ids:
            print(f"missing site IDs {missing_site_ids}")
            data = data.loc[~data["cpa_id"].isin(missing_site_ids), :]
        site_ids = data["cpa_id"].map(site_map).to_list()
    else:
        site_ids = [str(int(i)) for i in data["cpa_id"]]
    if profile_path is not None:
        cpa_profiles = load_site_profiles(profile_path, site_ids=list(set(site_ids)))
        profiles = [np.roll(cpa_profiles[site].values, utc_offset) for site in site_ids]
        data["profile"] = profiles

    # Split sites into bins using numeric features
    bin_features = []
    for b in bin or []:
        feature = snake_case_str(b.get("feature"))
        if not feature:
            raise KeyError(
                "One of your renewables_clusters uses the 'bin' option but doesn't include "
                "the 'feature' argument. You must specify a numeric feature (column) to "
                "split the renewable sites into bins."
            )
        if feature not in data.columns:
            raise KeyError(
                "One of your renewables_clusters uses the 'bin' option and includes the "
                f"feature argument '{feature}', which is not in the renewable site data. The "
                "feature must be one of the columns in your renewable site data file."
            )
        if not data[feature].dtype.kind in "iufc":
            raise TypeError(
                f"You specified the feature '{feature}' to bin one of your renewables_clusters. "
                f"'{feature}' is not a numeric column. Binning requires a numeric column."
            )

        weights_col = snake_case_str(b.get("weights"))
        if weights_col and weights_col not in data.columns:
            raise KeyError(
                "One of your renewables_clusters uses the 'bin' option and includes the "
                f"'weights' argument '{weights_col}', which is not in the renewable site data. The "
                "weights must be one of the columns in your renewable site data file.\n\n"
                "NOTE: Use the parameter 'mw' to weight by capacity."
            )
        elif weights_col:
            weights = data[weights_col]
        else:
            weights = None

        if not bin_features:
            b = num_bins_from_capacity(data, b)
            data[f"{feature}_bin"] = value_bin(
                data[feature], b.get("bins"), b.get("q"), weights=weights
            )
        else:
            df_list = []
            for _, _df in data.groupby(bin_features[-1]):
                _b = num_bins_from_capacity(_df, b.copy())
                _weights = None
                if weights is not None:
                    _weights = _df[weights_col]

                _df[f"{feature}_bin"] = value_bin(
                    _df[feature], _b.get("bins"), b.get("q"), weights=_weights
                )
                df_list.append(_df)
            data = pd.concat(df_list, ignore_index=True)
        bin_features.append(f"{feature}_bin")

    group_by = bin_features + ([snake_case_str(g) for g in group or []])
    prev_feature_cluster_col = None

    # Using a clustering method (agglomerative or kmeans)
    for clust in cluster or []:
        if isinstance(clust["feature"], str):
            clust["feature"] = snake_case_str(clust["feature"])
        else:
            clust["feature"] = [snake_case_str(f) for f in clust["feature"]]
        if "mw_per_cluster" in clust:
            if clust.get("n_clusters") is not None:
                logger.warning("Overwriting 'n_clusters' based on mw_cluster_size")
            if not group_by:
                clust["n_clusters"] = max(
                    int(data["mw"].sum() / clust["mw_per_cluster"]), 1
                )
            else:
                n_clusters = (
                    data.groupby(group_by)["mw"].sum() / clust["mw_per_cluster"]
                ).astype(int)
                clust["n_clusters"] = n_clusters.where(n_clusters > 0, 1)
            del clust["mw_per_cluster"]

        if "cluster" in data.columns and prev_feature_cluster_col:
            data = data.rename(columns={"cluster": prev_feature_cluster_col})
        if group_by:
            clust["by"] = group_by
            data = cluster_sites(True, data, **clust)
        else:
            data = cluster_sites(False, data, **clust)

        group_by.append(f"{clust['feature']}_clust")
        prev_feature_cluster_col = f"{clust['feature']}_clust"

    if "cluster" not in data.columns:
        if group_by:
            i = 1
            df_list = []
            for _, _df in data.groupby(group_by):
                _df["cluster"] = i
                df_list.append(_df)
                i += 1
            data = pd.concat(df_list)
        else:
            data["cluster"] = range(1, len(data) + 1)

    return data


def num_bins_from_capacity(data: pd.DataFrame, b: Dict[str, int]) -> Dict[str, int]:
    """Calculate the "bins" or "q" parameter based on available capacity.

    Either `mw_per_bin` or `mw_per_q` can be a key in the input dictionary. The
    number of bins/quantiles is calculated by dividing total site capacity by the
    mw_per_* value. If neither key is present, return the dictionary
    unaltered.

    Parameters
    ----------
    data : pd.DataFrame
        Must have column "mw"
    b : Dict[str, int]
        Must have either "mw_per_bin" or "mw_per_q" if the number of bins/quantiles
        should be decided based on available capacity.

    Returns
    -------
    Dict[str, int]
        Dictionary with either existing `bins`/`q` key or one of those keys calculated
        from available site capacity and `mw_per_bin`/`mw_per_q`
    """
    if "mw_per_bin" in b:
        if b.get("bins") is not None:
            logger.warning("Overwriting 'bins' based on mw_per_bin")
        b["bins"] = max(int(data["mw"].sum() / b["mw_per_bin"]), 1)
        del b["mw_per_bin"]
    elif "mw_per_q" in b:
        if b.get("q") is not None:
            logger.warning("Overwriting 'q' based on mw_per_q")
        b["q"] = max(int(data["mw"].sum() / b["mw_per_q"]), 1)
        del b["mw_per_q"]

    return b


def modify_renewable_group(
    df: pd.DataFrame, group_modifiers: List[Dict[str, Union[float, str, list]]] = None
) -> pd.DataFrame:
    """Modify values (e.g. cost) of a rewnewables cluster based on group membership.

    Parameters
    ----------
    df : pd.DataFrame
        Clustered renewable sites with averaged parameters like cost and profile. Must
        have column "cluster".
    group_modifiers : List[Dict[str, Union[float, str, list]]], optional
        List of dicts. Each must have keys "group" and "group_value". Any other keys should
         correspond to a column in `df` such as "capex_mw", "Inv_Cost_per_MW", etc. The
         values for these keys should either be a 2-item list with an operator and a value
         or a single numeric value.

         Rows of `df` are identified by string matching f"{group}:{group_value}" on the
         "cluster" column.

         By default None

    Returns
    -------
    pd.DataFrame
        Modified version of input df. No change in columns.

    Raises
    ------
    KeyError
        One dictionary in group_modifiers is missing either "group" or "group_value" keys
    ValueError
        The operator list is not a 2-item list. Must be 2 items (operator and value)
    ValueError
        The operator is not in the valid list (["add", "mul", "truediv", "sub"])
    """
    allowed_operators = ["add", "mul", "truediv", "sub"]
    for _group_mod in group_modifiers or []:
        group_mod = _group_mod.copy()
        missing_keys = [
            k for k in ["group", "group_value"] if k not in group_mod.keys()
        ]
        if missing_keys:
            raise KeyError(
                f"One of your 'renewables_clusters' has a 'group_modifiers' key but is "
                f"missing the key(s) {missing_keys}. These are required to modify values "
                "of a renewables cluster. Add these keys or remove the 'group_modifiers' "
                "section."
            )

        group = group_mod.pop("group")
        group_value = group_mod.pop("group_value")
        group_id = f"{group}_{group_value}"
        # for group, mod in (group_modifiers or {}).items():
        for key, op_list in group_mod.items():
            if isinstance(op_list, float) | isinstance(op_list, int):
                df.loc[
                    df["cluster"].str.contains(group_id, case=False, regex=False), key
                ] = op_list
            else:
                if len(op_list) != 2:
                    raise ValueError(
                        "Either a single numeric value or a list of two values - an operator "
                        "and a numeric value - are needed in the parameter "
                        f"'{key}' whenever 'group_modifiers' are used in 'renewables_clusters'. "
                        f"One of your 'group_modifiers' has {op_list} instead."
                    )
                op, op_value = op_list
                if op not in allowed_operators:
                    raise ValueError(
                        f"One of the 'group_modifiers' in your 'renewables_clusters' with key "
                        f"{key} has {op} as the mathmatical operator. Only {allowed_operators} "
                        "in the format [<operator>, <value>] can be used to modify the "
                        "properties of a renewable cluster.\n"
                    )
                f = operator.attrgetter(op)
                df.loc[
                    df["cluster"].str.contains(group_id, case=False, regex=False), key
                ] = f(operator)(
                    df.loc[
                        df["cluster"].str.contains(group_id, case=False, regex=False),
                        key,
                    ],
                    op_value,
                )
    return df



# Read in and add external inputs (user-supplied files) to PowerGenome outputs

import logging
from pathlib import Path
from typing import Any, List

import numpy as np
import pandas as pd

from powergenome.price_adjustment import inflation_price_adjustment
from powergenome.util import remove_feb_29

logger = logging.getLogger(__name__)


def make_demand_response_profiles(
    path: Path, resource_name: str, year: int, scenario: str
) -> pd.DataFrame:
    """Read files with DR profiles across years and scenarios. Return the hourly
    load profiles for a single resource in the model year.

    Parameters
    ----------
    path : path-like
        Where to load the file from
    resource_name : str
        Name of of the demand response resource
    year : int
        Year of data to use from the user demand response file
    scenario : str
        Name of scenario to use from the user demand response file

    Returns
    -------
    DataFrame
        Hourly profiles of DR load for each region where the resource is available.
        Column names are the regions plus 'scenario'.
    """

    df = pd.read_csv(path, header=[0, 1, 2, 3])

    # Use the MultiIndex columns to just get columns with the correct resource listed
    # in the top row of the csv. The resource name is dropped from the columns.
    resource_df = df.loc[:, resource_name]

    assert year in set(
        resource_df.columns.get_level_values(0).astype(int)
    ), f"The model year is not in the years of data for DR resource {resource_name}"

    resource_df = resource_df.loc[:, str(year)]

    assert scenario in set(
        resource_df.columns.get_level_values(0)
    ), f"The scenario {scenario} is not included for DR resource {resource_name}"

    resource_df = resource_df.loc[:, scenario]
    resource_df = resource_df.reset_index(drop=True)

    if len(resource_df) == 8784:
        remove_feb_29(resource_df)

    return resource_df


def demand_response_resource_capacity(df, resource_name, settings):
    """Calculate the maximum capacity value to assign a demand response/DSM resource

    Parameters
    ----------
    df : DataFrame
        Hourly demand profile in each region of a single DR resource
    resource_name : str
        Name of the resource, which should match the settings file
    settings : dict
        User-defined parameters from a settings file

    Returns
    -------
    DataFrame
        Index of scenarios and columns of regions, values represent shiftable capacity.
    """

    year = settings["model_year"]
    fraction_shiftable = settings["flexible_demand_resources"][year][resource_name][
        "fraction_shiftable"
    ]

    # peak_load = df.groupby(["scenario"]).max()
    peak_load = df.max()
    shiftable_capacity = peak_load * fraction_shiftable

    return shiftable_capacity


def add_resource_max_cap_spur(
    new_resource_df: pd.DataFrame, settings: dict, capacity_col: str = "Max_Cap_MW"
) -> pd.DataFrame:
    """Load user supplied maximum capacity and spur line data for new resources. Add
    those values to the resources dataframe.

    Parameters
    ----------
    new_resource_df : pd.DataFrame
        New resources that can be built. Each row should be a single resource, with
        columns 'region' and 'technology'. The number of copies for a region/resource
        (e.g. more than one UtilityPV resource in a region) should match what is
        given in the user-supplied file.
    settings : dict
        User-defined parameters from a settings file. Should have keys of `input_folder`
        (a Path object of where to find user-supplied data) and
        `capacity_limit_spur_fn` (the file to load).
    capacity_col : str, optional
        The column that indicates maximum capacity constraints, by default "Max_Cap_MW"

    Returns
    -------
    pd.DataFrame
        A modified version of new_resource_df with spur_miles and the maximum
        capacity for a resource. Copies of a resource within a region should have a
        cluster name to uniquely identify them.
    """

    defaults = {
        "cluster": 1,
        "spur_miles": 0,
        capacity_col: -1,
        "interconnect_annuity": 0,
        "interconnect_capex_mw": 0,
    }
    # Prepare file
    path = Path(settings["input_folder"]) / settings["capacity_limit_spur_fn"]
    df = pd.read_csv(path)
    if "technology" not in df.columns:
        raise KeyError("The max capacity/spur file must have column 'technology'")

    # Let users omit the "region" column or set a value of "all"
    if "region" not in df.columns:
        df["region"] = "all"
    for region in settings["model_regions"]:
        _df = df.loc[df["region"].str.lower() == "all", :]
        _df.loc[:, "region"] = region
        df = df.append(_df)

    df = df.loc[df["region"].str.lower() != "all", :]

    for key, value in defaults.items():
        df[key] = df[key].fillna(value) if key in df else value
        new_resource_df[key] = (
            new_resource_df[key].fillna(value) if key in new_resource_df else value
        )
    # Update resources
    grouped_df = df.groupby(["region", "technology"])
    for (region, tech), _df in grouped_df:
        mask = (new_resource_df["region"] == region) & (
            new_resource_df["technology"]
            .str.replace("_\*", "_all")
            .str.contains(tech.replace("_\*", "_all"), case=False, regex=False)
        )
        if mask.sum() > 1:
            resources = new_resource_df.loc[mask, "technology"].to_list()
            raise ValueError(
                f"Resource {tech} in region {region} from file "
                f"{settings['capacity_limit_spur_fn']} matches multiple resources:"
                f"\n{resources}"
            )
        else:
            pass
        for key, value in defaults.items():
            _key = "max_capacity" if key == capacity_col else key
            new_resource_df.loc[mask & (new_resource_df[key] == value), key] = _df[
                _key
            ].values
    logger.info(
        f"Inflating external interconnect annuity costs from 2017 to "
        f"{settings['target_usd_year']}"
    )
    new_resource_df["interconnect_annuity"] = inflation_price_adjustment(
        new_resource_df["interconnect_annuity"], 2017, settings["target_usd_year"]
    )
    return new_resource_df


def make_generator_variability(
    df: pd.DataFrame, remove_feb_29: bool = True
) -> pd.DataFrame:
    """Make a generator variability dataframe with normalized (0-1) hourly profiles
    for each resource in resource_df.

    Any resources that do not have a profile in column
    `profile` are assumed to have constant hourly profiles with a value of 1.
    February 29 is removed from any profiles of length 8784 (leap year).

    Parameters
    ----------
    df
        Dataframe with a single row for each resource. May have column `profile`.

    Returns
    -------
    pd.DataFrame
        A new dataframe with one column for each row of `df` and 8760 rows
        of generation profiles.

    Examples
    --------
    >>> df = pd.DataFrame({'profile': [[0] * 8760, np.ones(8784) / 2, None]})
    >>> make_generator_variability(df)
            0    1    2
    0     0.0  0.5  1.0
    1     0.0  0.5  1.0
    2     0.0  0.5  1.0
    3     0.0  0.5  1.0
    4     0.0  0.5  1.0
    ...   ...  ...  ...
    8755  0.0  0.5  1.0
    8756  0.0  0.5  1.0
    8757  0.0  0.5  1.0
    8758  0.0  0.5  1.0
    8759  0.0  0.5  1.0
    <BLANKLINE>
    [8760 rows x 3 columns]
    """

    def profile_len(x: Any) -> int:
        if isinstance(x, (list, np.ndarray)):
            return len(x)
        return 1

    def format_profile(
        x: Any, remove_feb_29: bool = True, hours: int = 8760
    ) -> np.ndarray:
        # from IPython import embed
        # embed()
        if isinstance(x, np.ndarray):
            if len(x) == 8784:
                # Remove February 29 from leap year
                if remove_feb_29:
                    return np.delete(x, slice(1416, 1440))
            return x
        if isinstance(x, list):
            return format_profile(np.array(x), remove_feb_29, hours)
        # Fill missing with default [1, ...]
        return np.ones(hours, dtype=float)

    if "profile" in df:
        hours = df["profile"].apply(profile_len).max()
        if remove_feb_29 and hours == 8784:
            hours = 8760
        kwargs = {"remove_feb_29": remove_feb_29, "hours": hours}
        profiles = np.column_stack(df["profile"].apply(format_profile, **kwargs).values)
        # Make sure values are not less than 0
        profiles = np.where(profiles >= 0, profiles, 0)
    # elif not remove_feb_29:
    #     profiles = np.ones((8760, len(df)), dtype=float)
    else:
        profiles = np.ones((8760, len(df)), dtype=float)
    return pd.DataFrame(profiles, columns=np.arange(len(df)).astype(str))


def load_policy_scenarios(settings: dict) -> pd.DataFrame:
    """Load the policy scenarios and copy cases where indicated. The policy file should
    start with columns `case_id` and `year`, and can contain an optional `copy_case_id`.
    Other columns should match the desired output format. The value `None` is included
    in na_values.

    Parameters
    ----------
    settings : dict
        User-defined parameters from a settings file. Should have keys of `input_folder`
        (a Path object of where to find user-supplied data) and
        `emission_policies_fn` (the file to load).

    Returns
    -------
    DataFrame
        Emission policies for each case_id/year.
    """

    path = Path(settings["input_folder"]) / settings["emission_policies_fn"]
    policies = pd.read_csv(path, na_values=["None", "none"])

    # Update the policies. The column `copy_case_id` can be used to copy values from
    # another policy to reduce human copy/paste errors.
    if "copy_case_id" in policies.columns:
        policies = copy_case_values(policies, match_cols=["case_id", "year", "region"])
        policies = policies.drop(columns="copy_case_id", errors="ignore")

    if "case_id" not in policies.columns:
        policies["case_id"] = settings["case_id"]

    warned = False
    if policies.duplicated(subset=["case_id", "year", "region"]).any():
        logger.warning(
            "Your emissions policies file has repeated values of 'case_id' and 'year'. "
            "This duplication of emission policies across multiple years may cause erors."
        )
        warned = True
    policies.loc[policies["case_id"].astype(str).str.lower() == "all", "case_id"] = (
        settings["case_id"]
    )
    if policies.duplicated(subset=["case_id", "year", "region"]).any() and not warned:
        logger.warning(
            f"After replacing values of 'all' with {settings['case_id']}, your emissions "
            "policies file has repeated values of 'case_id' and 'year'. "
            "This duplication of emission policies across multiple years may cause erors."
        )

    policies = policies.set_index(["case_id", "year"])

    return policies


def copy_case_values(df, match_cols):
    """Copy values for one case to others in an external data file. Must have column
    "copy_case_id"

    Parameters
    ----------
    df : DataFrame
        Policies, settings, or some other type of data for each case/year.
    match_cols : list
        The columns used to match cases. Must include "case_id", can also include things
        like "year" and "region".

    Returns
    -------
    DataFrame
        Modified copy of the original with values copied to rows with valid
        "copy_case_id" values.
    """
    match_cols_len = len(match_cols)
    settings_cols = [
        col for col in df.columns if col not in match_cols + ["copy_case_id"]
    ]
    for row in df.itertuples(index=False):
        if not pd.isna(row.copy_case_id):
            # Create a dictionary with keys of each col from match_cols and values
            # from the row. Then replace case_id key with copy_case_id value. This way
            # we get the new case id but same region, year, etc.
            filter_dict = {
                col: value
                for col, value in zip(df.columns[:match_cols_len], row[:match_cols_len])
            }
            filter_dict["case_id"] = row.copy_case_id  # Set to the case we want to copy

            # Use the filter dictionary to filter the original df and assign the values
            # from that row to the row in question.
            # https://stackoverflow.com/a/34162576
            new_values = df.loc[
                (df[list(filter_dict)] == pd.Series(filter_dict)).all(axis=1),
                settings_cols,
            ].values

            filter_dict["case_id"] = row.case_id  # Set back to case we want to modify

            df.loc[
                (df[list(filter_dict)] == pd.Series(filter_dict)).all(axis=1),
                settings_cols,
            ] = new_values

    return df


def load_demand_segments(settings):
    """Load a file of demand segments as defined by the user in CSV.

    Parameters
    ----------
    settings : dict
        User defined PowerGenome settings. Must have the keys "input_folder" and
        "demand_segments_fn".

    Returns
    -------
    DataFrame
        Demand segments with columns such as "Voll", "Demand_segment", etc.
    """

    path = Path(settings["input_folder"]) / settings["demand_segments_fn"]
    demand_segments = pd.read_csv(path)

    return demand_segments


def load_user_genx_settings(settings):
    """Load a file with user-supplied GenX settings and copy values to cases where
    needed.

    Parameters
    ----------
    settings : dict
        User defined PowerGenome settings. Must have the keys "input_folder" and
        "case_genx_settings_fn".

    Returns
    -------
    DataFrame
        Values for GenX settings across cases and years as defined by the user. The
        index is ["case_id", "year"] and columns are GenX parameters.
    """

    path = Path(settings["input_folder"]) / settings["case_genx_settings_fn"]
    genx_case_settings = pd.read_csv(path)

    if "copy_case_id" in genx_case_settings.columns:
        genx_case_settings = copy_case_values(
            genx_case_settings, match_cols=["case_id", "year"]
        )

    genx_case_settings = genx_case_settings.drop(columns=["copy_case_id"])
    genx_case_settings = genx_case_settings.set_index(["case_id", "year"])

    return genx_case_settings


def overwrite_wind_pv_capacity(df, settings):
    """Use external data to overwrite the wind and solarpv capacity extracted from
    EIA860.

    Parameters
    ----------
    df : DataFrame
        Existing generators dataframe, with columns "region", "technology", and
        "Existing_Cap_MW". The technologies should include "Solar Photovoltaic"
        and "Onshore Wind Turbine".
    settings : dict
        User defined PowerGenome settings. Must have the keys "input_folder" and
        "region_wind_pv_cap_fn".

    Returns
    -------
    DataFrame
        Same as input dataframe but with new capacity values for technologies defined
        in the "region_wind_pv_cap_fn" file.
    """
    from powergenome.util import reverse_dict_of_lists

    idx = pd.IndexSlice

    path = settings["input_folder"] / settings["region_wind_pv_cap_fn"]

    wind_pv_ipm_region_capacity = pd.read_csv(path)

    region_agg_map = reverse_dict_of_lists(settings.get("region_aggregations", {}))

    # Set model_region as IPM_region to start
    wind_pv_ipm_region_capacity["model_region"] = wind_pv_ipm_region_capacity[
        "IPM_Region"
    ]
    # Change any aggregated regions to the user-defined model_region
    wind_pv_ipm_region_capacity.loc[
        wind_pv_ipm_region_capacity["IPM_Region"].isin(region_agg_map), "model_region"
    ] = wind_pv_ipm_region_capacity.loc[
        wind_pv_ipm_region_capacity["IPM_Region"].isin(region_agg_map), "IPM_Region"
    ].map(
        region_agg_map
    )
    wind_pv_model_region_capacity = wind_pv_ipm_region_capacity.groupby(
        ["model_region", "technology"]
    ).sum()

    df = df.reset_index()

    for region in df["region"].unique():
        for tech in ["Solar Photovoltaic", "Onshore Wind Turbine"]:
            if tech in df.query("region == @region")["technology"].to_list():
                try:
                    df.loc[
                        (df["region"] == region) & (df["technology"] == tech),
                        "Existing_Cap_MW",
                    ] = wind_pv_model_region_capacity.loc[
                        idx[region, tech], "nameplate_capacity_mw"
                    ]
                except:
                    pass

    df = df.set_index(["region", "technology", "cluster"])

    return df


def make_usr_demand_profiles(path, settings):
    idx = pd.IndexSlice
    year = settings["model_year"]
    scenario = settings.get("electrification")
    if not scenario:
        scenario_file = settings.get("scenario_definitions_fn")
        load_file = settings.get("regional_load_fn")
        raise KeyError(
            f"The scenario definitions file {scenario_file} must have a column "
            f"'electrification' with values that correspond to the second row of "
            f"{load_file}"
        )

    df = pd.read_csv(path, header=[0, 1, 2])
    scenario_df = df.loc[:, idx[str(year), scenario]]

    return scenario_df


def load_user_tx_costs(
    path: Path, model_regions: List[str], target_usd_year: int = None
) -> pd.DataFrame:
    """Load a user data file with cost and line loss of each interregional transmission
    line. Map the region names to zones (z1 to zM) and adjust the total cost columns
    to the dollar year specified in settings.

    Parameters
    ----------
    path : Path
        Path to the CSV file, which should have columns "start_region", "dest_region",
        "total_interconnect_annuity_mw", "total_interconnect_cost_mw", and "dollar_year".
    model_regions : List[str]
        List of model region names. Should be sorted to match order in other functions.
    target_usd_year : int, optional
        Desired final dollar year of cost columns, by default None. If None, no adjustment
        is made.

    Returns
    -------
    pd.DataFrame
        Cost and line loss of each potential transmission line between model regions.
        Contains columns "start_region", "dest_region", "zone_1", "zone_2",
        "total_interconnect_annuity_mw", "total_interconnect_cost_mw", "dollar_year",
        and "adjusted_dollar_year".
    """
    df = pd.read_csv(path)

    zones = model_regions
    zone_num_map = {
        zone: f"z{number + 1}" for zone, number in zip(zones, range(len(zones)))
    }
    df["zone_1"] = df["start_region"].map(zone_num_map)
    df["zone_2"] = df["dest_region"].map(zone_num_map)
    df = df.dropna(subset=["zone_1", "zone_2"])

    if target_usd_year:
        adjusted_annuities = []
        adjusted_costs = []
        for row in df.itertuples():
            adj_annuity = inflation_price_adjustment(
                row.total_interconnect_annuity_mw, row.dollar_year, target_usd_year
            ).round(0)
            adjusted_annuities.append(adj_annuity)

            adj_cost = inflation_price_adjustment(
                row.total_interconnect_cost_mw, row.dollar_year, target_usd_year
            ).round(0)
            adjusted_costs.append(adj_cost)
        df["total_interconnect_annuity_mw"] = adjusted_annuities
        df["total_interconnect_cost_mw"] = adjusted_costs
        df["adjusted_dollar_year"] = target_usd_year

    return df


def insert_user_tx_costs(tx_df: pd.DataFrame, user_costs: pd.DataFrame) -> pd.DataFrame:
    """Insert costs and line loss from the user. Can include more lines than were in the
    original transmission dataframe to create lines with zero existing capacity.

    Parameters
    ----------
    tx_df : pd.DataFrame
        Dataframe of interregional transmission lines created by PG. Must have the column
        "Network_Lines" (integer from 1 to N), a column for each network zone (z1 to zM),
        "Line_Max_Flow_MW", "Line_Min_Flow_MW", and "transmission_path_name".
    user_costs : pd.DataFrame
        Dataframe of costs and line loss created by the user. Should have columns
        "zone_1" and "zone_2" (possible values are z1 through zM), "total_interconnect_annuity_mw",
        "total_interconnect_cost_mw", and "total_line_loss_frac". Cost values should
        already be adjusted to the desired dollar year.

    Returns
    -------
    pd.DataFrame
        Supplemented interregional transmission lines
    """
    if tx_df.empty:
        return tx_df
    user_costs = user_costs.dropna(subset=["zone_1", "zone_2"], how="any")
    unused_lines = []
    for row in user_costs.itertuples():
        line_row = tx_df.loc[(tx_df[row.zone_1] != 0) & (tx_df[row.zone_2] != 0), :]
        assert not len(line_row) > 1

        if line_row.empty:
            unused_lines.append(row)
        tx_df.loc[line_row.index, "Line_Reinforcement_Cost_per_MWyr"] = (
            row.total_interconnect_annuity_mw
        )
        tx_df.loc[line_row.index, "Line_Reinforcement_Cost_per_MW"] = (
            row.total_interconnect_cost_mw
        )
        tx_df.loc[line_row.index, "Line_Loss_Percentage"] = row.total_line_loss_frac

    unused_line_df = pd.DataFrame(unused_lines)
    unused_line_df = unused_line_df.rename(
        columns={
            "total_interconnect_annuity_mw": "Line_Reinforcement_Cost_per_MWyr",
            "total_interconnect_cost_mw": "Line_Reinforcement_Cost_per_MW",
            "total_line_loss_frac": "Line_Loss_Percentage",
        }
    )

    zone_cols = [c for c in tx_df.columns if c[0] == "z"]
    unused_line_df[zone_cols] = 0
    for idx, row in unused_line_df.iterrows():
        unused_line_df.loc[idx, row["zone_1"]] = 1
        unused_line_df.loc[idx, row["zone_2"]] = -1
        unused_line_df.loc[idx, "transmission_path_name"] = (
            f"{row.start_region}_to_{row.dest_region}"
        )

    cols = [c for c in tx_df.columns if c in unused_line_df.columns]
    tx_df = pd.concat([tx_df, unused_line_df[cols]], ignore_index=True)
    tx_df["Network_Lines"] = range(1, len(tx_df) + 1)
    tx_df["Line_Max_Flow_MW"] = tx_df["Line_Max_Flow_MW"].fillna(0)
    tx_df["Line_Min_Flow_MW"] = tx_df["Line_Max_Flow_MW"].fillna(0)

    return tx_df



"""
Parameters and settings
"""

import os
from pathlib import Path
from typing import Union

from dotenv import load_dotenv

from powergenome import __file__
from powergenome.resource_clusters import ClusterBuilder
from powergenome.util import sqlalchemy_prefix

# Not convinced this is the best way to set folder paths but it works!
powergenome_path = Path(__file__).parent
project_path = powergenome_path.parent

load_dotenv(dotenv_path=powergenome_path / ".env")


DATA_PATHS = {}
DATA_PATHS["results"] = project_path / "results"
DATA_PATHS["powergenome"] = project_path / "powergenome"
DATA_PATHS["data"] = project_path / "data"
DATA_PATHS["atb_storage_costs"] = DATA_PATHS["data"] / "NREL_ATB_battery_costs.csv"
DATA_PATHS["ipm_shapefiles"] = DATA_PATHS["data"] / "IPM Regions v617 04-05-17"
DATA_PATHS["tests"] = project_path / "tests"
DATA_PATHS["test_data"] = DATA_PATHS["tests"] / "data"
DATA_PATHS["settings"] = project_path / "settings"
DATA_PATHS["eia"] = DATA_PATHS["data"] / "eia"
DATA_PATHS["eia_860m"] = DATA_PATHS["eia"] / "860m"
DATA_PATHS["cost_multipliers"] = DATA_PATHS["data"] / "cost_multipliers"
DATA_PATHS["additional_techs"] = DATA_PATHS["data"] / "additional_technologies"
DATA_PATHS["coal_fgd"] = DATA_PATHS["data"] / "coal_fgd" / "fgd_output.csv"
DATA_PATHS["cpi_data"] = DATA_PATHS["data"] / "cpi_data" / "cpi_data.csv"

IPM_SHAPEFILE_PATH = DATA_PATHS["ipm_shapefiles"] / "IPM_Regions_201770405.shp"
IPM_GEOJSON_PATH = DATA_PATHS["data"] / "ipm_regions_simple.geojson"

SETTINGS = {}
SETTINGS["PUDL_DB"] = sqlalchemy_prefix(os.environ.get("PUDL_DB"))
SETTINGS["PG_DB"] = sqlalchemy_prefix(os.environ.get("PG_DB"))
SETTINGS["EFS_DATA"] = os.environ.get("EFS_DATA")
SETTINGS["RESOURCE_GROUPS"] = os.environ.get("RESOURCE_GROUPS")
SETTINGS["DISTRIBUTED_GEN_DATA"] = os.environ.get("DISTRIBUTED_GEN_DATA")
SETTINGS["RESOURCE_GROUP_PROFILES"] = os.environ.get("RESOURCE_GROUP_PROFILES")


def build_resource_clusters(
    group_path: Union[str, Path] = None, profile_path: Union[str, Path] = None
) -> ClusterBuilder:
    if not group_path:
        group_path = SETTINGS.get("RESOURCE_GROUPS")
    if not profile_path:
        profile_path = SETTINGS.get("RESOURCE_GROUP_PROFILES")
    if profile_path is not None:
        profile_path = Path(profile_path)
    if not group_path:
        cluster_builder = ClusterBuilder([])
    else:
        cluster_builder = ClusterBuilder.from_json(
            Path(group_path, ".").glob("**/*.json"), profile_path
        )
    return cluster_builder



import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
import pyarrow.parquet as pq

from powergenome.load_construction import us_state_abbrev
from powergenome.params import SETTINGS
from powergenome.util import deep_freeze_args, snake_case_col

logger = logging.getLogger(__name__)


def load_region_pop_frac(
    path_in: Path,
    fn: str = "ipm_state_pop_weight_20220329.csv",
) -> pd.DataFrame:
    if (path_in / fn).suffix == ".csv":
        pop = pd.read_csv(path_in / fn)
    elif (path_in / fn).suffix == ".parquet":
        pop = pd.read_parquet(path_in / fn)
    pop_cols = ["region", "dg_region", "frac_dg_in_region"]
    return pop[pop_cols]


def interp_dg(df: pd.DataFrame, year1: int, year2: int, target_year: int) -> pd.Series:
    """Linear interpolation of distributed generation between two data years

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with columns "time_index", "year", and "region_distpv_mwh"
    year1 : int
        First year of data in the df
    year2 : int
        Second year of data in the df
    target_year : int
        The interpolation target year

    Returns
    -------
    pd.Series
        Interpolated values.
    """
    if year1 > year2:
        year1, year2 = year2, year1

    s1 = df.loc[df["year"] == year1, :].set_index("time_index")["region_distpv_mwh"]
    s2 = df.loc[df["year"] == year2, :].set_index("time_index")["region_distpv_mwh"]

    if year1 < target_year < year2:
        w1 = 1 / abs(target_year - year1)
        w2 = 1 / abs(year2 - target_year)
        return ((s1 * w1) + (s2 * w2)) / (w1 + w2)
    elif year1 == target_year:
        return s1
    elif year2 == target_year:
        return s2


@deep_freeze_args
@lru_cache()
def distributed_gen_profiles(
    profile_fn: str,
    year: int,
    scenario: str,
    regions: List[str],
    path_in: Path,
    region_aggregations: Dict[str, str] = None,
    tz_offset: int = None,
) -> pd.DataFrame:
    """Build distributed generation profiles for each region

    Parameters
    ----------
    profile_fn : str
        Name of the data file with distributed generation profiles
    year : int
        Model planning year
    scenario : str
        Scenario name that matches a value from the data file
    regions : List[str]
        Base model regions (not aggregated). These should match the population weight file,
        not the regions in the data file.
    path_in : Path
        Folder where the distributed generation data file and population weight file are
        found.
    region_aggregations : Dict[str, str]
        Mapping of region aggregations from base regions to model regions (where used),
        by default None
    tz_offset: int
        Number of hours offset from the original data, by default None. If None, the
        original data will not be modified.

    Returns
    -------
    pd.DataFrame
        Wide dataframe with hourly distributed generation from each model region

    Raises
    ------
    KeyError
        No path_in was provided with the DISTRIBUTED_GEN_DATA parameter in the settings
        file or the .env file
    KeyError
        None of the base regions were found in the population weight file
    KeyError
        The distributed generation data file is missing required columns
    ValueError
        The electrification_scenario parameter does not match the values in the data file
    """
    if not path_in:
        try:
            path_in = Path(SETTINGS["DISTRIBUTED_GEN_DATA"])
        except TypeError:
            raise KeyError(
                "The variable 'DISTRIBUTED_GEN_DATA' is not included in your .env file."
            )
    else:
        path_in = Path(path_in)

    pop_files = path_in.glob("*pop_weight*")
    newest_pop_file = max(pop_files, key=os.path.getctime)
    pop = load_region_pop_frac(path_in=path_in, fn=newest_pop_file.name)
    pop = pop.loc[pop["region"].isin(regions), :]

    # NREL stores their state names as abbreviations. Change to full state name if found.
    if any([p in us_state_abbrev.keys() for p in pop["dg_region"]]):
        for dg_reg in pop["dg_region"].unique():
            pop.loc[pop["dg_region"] == dg_reg, "dg_region"] = us_state_abbrev[dg_reg]

    dg_regions = pop["dg_region"].unique()
    if dg_regions.size == 0:
        raise KeyError(
            "None of your regions were found in the population weighting file "
            f"({newest_pop_file})."
        )

    # Check file to identify possible errors in advance
    pf = pq.ParquetFile(path_in / profile_fn)
    col_names = pf.schema.names
    missing_names = []
    for col in ["region", "time_index", "year", "scenario", "distpv_MWh"]:
        if col.lower() not in [c.lower() for c in col_names]:
            missing_names.append(col)
    if missing_names:
        raise KeyError(
            f"The required columns {missing_names} are not in your distributed generation file"
        )

    scenario_profile = pd.read_parquet(
        path_in / profile_fn,
        filters=[("region", "in", dg_regions), ("scenario", "==", scenario)],
    )
    scenario_profile.columns = snake_case_col(scenario_profile.columns)
    if scenario_profile.empty:
        full_df = pd.read_parquet(
            path_in / profile_fn,
        )
        valid_scenarios = list(full_df.scenario.unique())
        if scenario not in valid_scenarios:
            raise ValueError(
                f"Your 'distributed_gen_scenario' parameter value '{scenario}' was not "
                f"found in the 'distributed_gen_fn' file '{profile_fn}'. Valid "
                f"scenarios are: {valid_scenarios}."
            )
    if not set(dg_regions) == set(scenario_profile["region"].unique()):
        missing_regions = set(dg_regions) - set(scenario_profile["region"].unique())
        logger.warning(
            f"No distributed generation for the regions {missing_regions} was found in "
            f"your distributed generation data for the scenario {scenario}. This is an "
            "informational warning -- check your data file if you think there should be "
            "distributed generation data in this/these region(s)."
        )
    # NREL data are available every other year.
    dg_years = scenario_profile.year.unique()
    if year in dg_years:
        years = [year]
    elif year not in dg_years and (min(dg_years) < year < max(dg_years)):
        year1 = max([y for y in dg_years if y < year])
        year2 = min([y for y in dg_years if y > year])
        years = [year1, year2]
        logger.info(
            f"Distributed generation profiles are not available for the year {year}. "
            f"Using a linear interpolation between {year1} and {year2}."
        )
    elif year not in dg_years and all(year < dg_years):
        years = [min(dg_years)]
        logger.warning(
            f"The first data year for distributed generation profiles is {years[0]}, "
            f"which is after your planning period year ({year}). Using the first data "
            "year since no earlier data have been provided."
        )
    else:
        years = [max(dg_years)]
        logger.warning(
            f"The last data year for distributed generation profiles is {years[0]}, "
            f"which is before your planning period year ({year}). Using the last data "
            "year since no later data have been provided."
        )
    scenario_profile = scenario_profile.loc[
        scenario_profile["year"].isin(years), :
    ].rename(columns={"region": "dg_region"})
    scenario_profile = pd.merge(scenario_profile, pop, on=["dg_region"])

    if not set(regions) == set(scenario_profile["region"].unique()):
        missing_regions = set(regions) - set(scenario_profile["region"].unique())
        logger.warning(
            f"No distributed generation for the regions {missing_regions} was found in "
            f"your distributed generation data for the scenario {scenario}. This is an "
            "informational warning -- check your data file if you think there should be "
            "distributed generation data in this/these region(s)."
        )

    for k, v in (region_aggregations or {}).items():
        scenario_profile.loc[scenario_profile["region"].isin(v), "region"] = k

    scenario_profile["region_distpv_mwh"] = (
        scenario_profile.distpv_mwh * scenario_profile.frac_dg_in_region
    )
    if len(years) == 2:
        # First combine by year/region. A region might be represented more than once.
        region_scenario_year_profile = scenario_profile.groupby(
            ["year", "region", "time_index"], as_index=False
        )["region_distpv_mwh"].sum()

        # Then interpolate between data years
        region_scenario_profile = (
            region_scenario_year_profile.groupby("region")
            .apply(interp_dg, years[0], years[1], year)
            .T
        )
    else:
        region_scenario_profile = scenario_profile.groupby(
            ["time_index", "region"], as_index=True
        )["region_distpv_mwh"].sum()
        region_scenario_profile = region_scenario_profile.unstack()

    # Adjust for timezone
    if tz_offset is not None:
        for col in region_scenario_profile.columns:
            region_scenario_profile[col] = np.roll(
                region_scenario_profile[col], tz_offset
            )
    return region_scenario_profile.reset_index(drop=True)



"Load data from file for pipeline/transport/injection costs"

import logging
from pathlib import Path
from typing import Dict, List

import pandas as pd

from powergenome.price_adjustment import inflation_price_adjustment
from powergenome.util import snake_case_str

logger = logging.getLogger(__name__)


def merge_co2_pipeline_costs(
    df: pd.DataFrame,
    co2_data_path: Path,
    co2_pipeline_filters: List[dict],
    region_aggregations: Dict[str, List[str]] = None,
    fuel_emission_factors: Dict[str, float] = None,
    extra_ccs_cost_tonne: float = None,
    target_usd_year: int = None,
) -> pd.DataFrame:
    """Merge columns with CO2 pipeline costs to a dataframe.

    Data come from an external file. Filters from the settings are used to select the
    data that will be merged.

    An example of the YAML settings parameter would be:

    co2_pipeline_filters:
      - technology: NaturalGas
        tech_detail: CCCCSAvgCF
        with_backbone: false
        percentile: 25

    Parameters
    ----------
    df : pd.DataFrame
        Generators data. Should include the rows "technology" and "region", along with a
        heat rate column that has "heat_rate" and "per_mwh" if fuel emission factors
        will be used to convert costs from per tonne to per MWh.
    co2_data_path : Path
        Path to the file with CO2 pipeline cost data
    co2_pipeline_filters : List[dict]
        Filters to apply to the CO2 cost data file
    region_aggregations : Dict[str, List[str]], optional
        Aggregations of multiple base regions to model regions, by default None
    fuel_emission_factors : Dict[str, float], optional
        CO2 emissions per unit of energy for different fuel types, by default None
    extra_ccs_cost_tonne: float
        An extra cost (positive or negative) assigned per tonne of CO2. Converted to
        cost per MWh.
    target_usd_year : int, optional
        The target year for cost inflation adjustment, by default None

    Returns
    -------
    pd.DataFrame
        The input dataframe (one row per resource) plus columns with cost data for CO2
        pipeline constuction/operation and CO2 disposal
    """
    logger.info("Adding CCS pipeline costs")
    co2_df = pd.read_csv(co2_data_path)
    co2_df = co2_df.loc[co2_df["parameter"] != "capacity_mw", :]
    if target_usd_year:
        for dollar_year in co2_df["dollar_year"].unique():
            co2_df.loc[(co2_df["dollar_year"] == dollar_year), "parameter_value"] = (
                inflation_price_adjustment(
                    co2_df.loc[co2_df["dollar_year"] == dollar_year, "parameter_value"],
                    dollar_year,
                    target_usd_year,
                )
            )
    for k, v in (region_aggregations or {}).items():
        co2_df.loc[co2_df["region"].isin(v), "region"] = k

    # When regions are aggregated, use the lowest-cost ipm region
    co2_df = co2_df.sort_values("parameter_value").drop_duplicates(
        [
            "region",
            "technology",
            "tech_detail",
            "with_backbone",
            "percentile",
            "parameter",
        ],
        keep="first",
    )

    # co2_df["resource"] = (co2_df["technology"] + "_" + co2_df["tech_detail"]).lower()
    filter_techs = [
        filt["technology"] + "_" + filt["tech_detail"] for filt in co2_pipeline_filters
    ]
    co2_resources = {}
    for co2_r in filter_techs:
        for model_r in df["technology"].unique():
            if snake_case_str(co2_r) in snake_case_str(model_r):
                co2_resources[co2_r] = model_r

    if not co2_resources:
        logger.warning(
            "You have specified technologies for CO2 pipeline costs in the parameter "
            "'co2_pipeline_filters' but none of the technologies are being used in "
            "your system."
        )
        return df

    df_list = []
    for filt in co2_pipeline_filters:
        # tech = filt["technology"].lower() + "_" + filt["tech_detail"].lower()
        query_str = f"technology == '{filt['technology']}'"
        for k, v in filt.items():
            if k != "technology":
                if isinstance(v, str):
                    query_str += f" and {k} == '{v}'"
                else:
                    query_str += f" and {k} == {v}"

        co2_costs = co2_df.query(query_str)
        co2_costs.loc[:, "technology"] = (
            co2_costs["technology"] + "_" + co2_costs["tech_detail"]
        ).map(co2_resources)
        co2_costs_wide = co2_costs.pivot(
            index=["region", "technology"],
            columns="parameter",
            values="parameter_value",
        ).reset_index()
        co2_costs_wide["capture_rate"] = filt.get("capture_rate", 90)
        df_list.append(co2_costs_wide)

    full_co2_costs_wide = pd.concat(df_list)
    full_co2_costs_wide["extra_ccs_cost_tonne"] = extra_ccs_cost_tonne or 0
    # Split incoming df into techs with and without pipeline cost. Remove pipeline cost
    # columns from the "with" techs to avoid duplicates, merge in co2 costs, then concat
    # with the "without" techs
    df = df.reset_index()
    drop_cols = [
        c for c in full_co2_costs_wide.columns if c not in ["region", "technology"]
    ]
    df_co2_techs = df.loc[df["technology"].isin(co2_resources.values()), :]
    df_co2_techs = df_co2_techs.drop(columns=drop_cols, errors="ignore")
    df_co2_techs = pd.merge(
        df_co2_techs, full_co2_costs_wide, on=["region", "technology"], how="left"
    )
    df_non_co2_techs = df.loc[~df["technology"].isin(co2_resources.values()), :]

    df_co2_costs = (
        pd.concat([df_co2_techs, df_non_co2_techs]).set_index("index").sort_index()
    )

    drop_idx = df_co2_costs.loc[
        (df_co2_costs["technology"].isin(co2_resources.values()))
        & (df_co2_costs[co2_costs_wide.columns].isna().any(axis=1))
    ].index

    if not drop_idx.empty:
        for tech, _df in df_co2_costs.loc[drop_idx, :].groupby("technology"):
            regs = _df["region"].to_list()
            logger.warning(
                f"The CCS resource {tech} is being removed from regions {regs} because cost "
                "data was not included in your CO2 pipeline cost file."
            )
    df_co2_costs = df_co2_costs.drop(index=drop_idx)
    df_co2_costs.loc[:, co2_costs_wide.columns] = df_co2_costs[
        co2_costs_wide.columns
    ].fillna(0)

    mass_cols = [c for c in full_co2_costs_wide.columns if "tonne" in c]
    if mass_cols and fuel_emission_factors:
        df_co2_costs = mass_to_energy_costs(
            df_co2_costs, mass_cols, fuel_emission_factors
        )

    return df_co2_costs


def mass_to_energy_costs(
    df: pd.DataFrame,
    mass_cols: List[str],
    fuel_emission_factors: Dict[str, float],
    heat_rate_col: str = None,
) -> pd.DataFrame:
    """Convert costs per tonne of CO2 to a MWh basis

    Parameters
    ----------
    df : pd.DataFrame
        Generators dataframe with a fuel column, a heat rate column, and the columns
        provided in the list `mass_cols`
    mass_cols : List[str]
        Names of columns with cost per tonne of CO2 disposal
    fuel_emission_factors : Dict[str, float]
        Emission rate of tonne per unit of energy for fuels used by CCS generators
    heat_rate_col : str, optional
        Name of the column with generator heat rates, by default None

    Returns
    -------
    pd.DataFrame
        Modified copy of the input, with the new column "co2_cost_mwh"
    """
    if not heat_rate_col:
        heat_rate_cols = [
            c for c in df.columns if "heat_rate" in c.lower() and "per_mwh" in c.lower()
        ]
        if len(heat_rate_cols) > 1:
            logger.warning(
                "More than one column with 'heat_rate' and 'per_mwh' was found when "
                "converting CO2 transport costs from per tonne to per MWh. Values are "
                "not being converted."
            )
            return df

        elif not heat_rate_cols:
            logger.warning(
                "Unable to identify a heat rate column when converting CO2 transport costs "
                "from per tonne to per MWh. A heat rate column should contain 'heat_rate' and "
                "'per_mwh' (capitalization does not matter). Values are not being converted."
            )
            return df
        else:
            heat_rate_col = heat_rate_cols[0]

    fuel_col = [c for c in df.columns if c.lower() == "fuel"]
    fuel_col = fuel_col[0]
    for fuel, ef in fuel_emission_factors.items():
        df.loc[
            df[fuel_col].str.contains(fuel, case=False, regex=False),
            "tonne_co2_captured_mwh",
        ] = (
            df.loc[
                df[fuel_col].str.contains(fuel, case=False, regex=False), heat_rate_col
            ]
            * ef
            * (
                df.loc[
                    df[fuel_col].str.contains(fuel, case=False, regex=False),
                    "capture_rate",
                ]
                / 100
            )
        )

    df.loc[:, "co2_cost_mwh"] = df[mass_cols].sum(axis=1) * df["tonne_co2_captured_mwh"]
    df.loc[:, "co2_cost_mwh"] = df.loc[:, "co2_cost_mwh"].fillna(0)

    return df



import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

from powergenome.params import SETTINGS
from powergenome.util import deep_freeze_args, find_region_col, snake_case_col

logger = logging.getLogger(__name__)


us_state_abbrev = {
    "Alabama": "AL",
    "Alaska": "AK",
    "American Samoa": "AS",
    "Arizona": "AZ",
    "Arkansas": "AR",
    "California": "CA",
    "Colorado": "CO",
    "Connecticut": "CT",
    "Delaware": "DE",
    "District Of Columbia": "DC",
    "District of Columbia": "DC",
    "Florida": "FL",
    "Georgia": "GA",
    "Guam": "GU",
    "Hawaii": "HI",
    "Idaho": "ID",
    "Illinois": "IL",
    "Indiana": "IN",
    "Iowa": "IA",
    "Kansas": "KS",
    "Kentucky": "KY",
    "Louisiana": "LA",
    "Maine": "ME",
    "Maryland": "MD",
    "Massachusetts": "MA",
    "Michigan": "MI",
    "Minnesota": "MN",
    "Mississippi": "MS",
    "Missouri": "MO",
    "Montana": "MT",
    "Nebraska": "NE",
    "Nevada": "NV",
    "New Hampshire": "NH",
    "New Jersey": "NJ",
    "New Mexico": "NM",
    "New York": "NY",
    "North Carolina": "NC",
    "North Dakota": "ND",
    "Northern Mariana Islands": "MP",
    "Ohio": "OH",
    "Oklahoma": "OK",
    "Oregon": "OR",
    "Pennsylvania": "PA",
    "Puerto Rico": "PR",
    "Rhode Island": "RI",
    "South Carolina": "SC",
    "South Dakota": "SD",
    "Tennessee": "TN",
    "Texas": "TX",
    "Utah": "UT",
    "Vermont": "VT",
    "Virgin Islands": "VI",
    "Virginia": "VA",
    "Washington": "WA",
    "West Virginia": "WV",
    "Wisconsin": "WI",
    "Wyoming": "WY",
}


def load_region_pop_frac(
    path_in: Path,
    fn: str = "ipm_state_pop_weight.parquet",
) -> pd.DataFrame:
    # TODO #178 finalize state pop weight file and filename
    # read in state proportions
    # how much state load should be distributed to GenXRegion

    if (path_in / fn).suffix == ".csv":
        pop = pd.read_csv(path_in / fn)
    elif (path_in / fn).suffix == ".parquet":
        pop = pd.read_parquet(path_in / fn)
    pop["state"] = pop["state"].map(us_state_abbrev)
    context = "Loading region population fraction file for EFS load construction."
    region_col = find_region_col(pop.columns, context)
    pop = pop.rename(columns={region_col: "region"})
    pop_cols = ["region", "state", "state_prop"]
    return pop[pop_cols]


running_sectors = {
    "res_space_heat_cool": ("Residential", "space heating and cooling"),
    "res_water_heat": ("Residential", "water heating"),
    "comm_space_heat_cool": ("Commercial", "space heating and cooling"),
    "comm_water_heat": ("Commercial", "water heating"),
    "trans_light_duty": ("Transportation", "light-duty vehicles"),
    "trans_medium_duty": ("Transportation", "medium-duty trucks"),
    "trans_heavy_duty": ("Transportation", "heavy-duty trucks"),
    "trans_bus": ("Transportation", "transit buses"),
}


def CreateOutputFolder(case_folder):
    path = case_folder / "extra_outputs"
    if not os.path.exists(path):
        os.makedirs(path)


def create_subsector_ts(
    sector: str,
    subsector: str,
    year: int,
    scenario_stock: pd.DataFrame,
    utc_offset: int = 0,
    path_in: Path = None,
) -> pd.DataFrame:
    """Build hourly demand for a single sector/subsector in a given year and scenario.

    This uses files with stock projections and "incremental factors" of hourly demand
    for each stock type. Two stock types and hourly factors are supported (based on how
    data were originally derived from NREL EFS). Data can be shifted from the original
    timezone (e.g. UTC) to a different timezone.

    Sector and subsector names must correspond to the stock and incremental factor files
    used.

    Stock values must be for the model year specified.

    Parameters
    ----------
    sector : str
        Name of the stock sector (e.g. Residential)
    subsector : str
        Name of the stock subsector (e.g. water heating)
    year : int
        Modeling year, for selecting incremental factors
    scenario_stock : pd.DataFrame
        Stock of different sector/subsectors in the model year in applicable states. Must
        have the columns "sector", "subsector", "state", "year", "agg_stock_type1", and
        "agg_stock_type2".
    utc_offset : int, optional
        Number of hours to shift the data away from UTC, by default 0
    path_in : Path, optional
        Folder with incremental factor timeseries data, by default None. There should be
        a file in this folder named with the convention "{sector}_{subsector}_Incremental_Factor.parquet"

    Returns
    -------
    pd.DataFrame
        Hourly demand of a single sector/subsector across one or more states, adjusted
        to the specified timezone (relative to data storage)
    """
    if not path_in:
        try:
            path_in = Path(SETTINGS["EFS_DATA"])
        except TypeError:
            logger.warning("The variable 'EFS_DATA' is not included in your .env file.")

    timeseries = pd.read_parquet(
        path_in / f"{sector}_{subsector}_Incremental_Factor.parquet"
    )
    timeseries.columns = snake_case_col(timeseries.columns)
    timeseries = timeseries.rename(columns={"localhourid": "time_index"})
    ts_cols = ["state", "year", "time_index", "unit", "factor_type1", "factor_type2"]
    timeseries = timeseries.loc[:, ts_cols]
    stock_temp = scenario_stock.loc[
        (scenario_stock["sector"] == sector)
        & (scenario_stock["subsector"] == subsector),
        ["state", "year", "agg_stock_type1", "agg_stock_type2"],
    ]

    factor_years = timeseries["year"].unique()
    if year not in factor_years:
        diff = np.array(factor_years - year)
        index = diff[np.where(diff <= 0)].argmax()
        year_approx = factor_years[index]
        timeseries = timeseries.loc[timeseries["year"] == year_approx, :]
        timeseries["year"] = year
        logger.warning(
            f"No incremental factor available for {sector}/{subsector} "
            f"in year {year}: using factors from year {year_approx} instead."
        )

    timeseries = pd.merge(timeseries, stock_temp, on=["state", "year"], validate="m:1")
    timeseries = timeseries.assign(
        load_mw=timeseries["agg_stock_type1"] * timeseries["factor_type1"]
        + timeseries["agg_stock_type2"] * timeseries["factor_type2"]
    )
    timeseries = (
        timeseries[["state", "time_index", "load_mw"]]
        .dropna()
        .groupby("state")
        .apply(utc_offset_state_load, utc_offset)
    )
    return timeseries


def utc_offset_state_load(df: pd.DataFrame, utc_offset: int = 0) -> pd.DataFrame:
    """Shift hourly load within a single dataframe from UTC to an offset timezone

    Parameters
    ----------
    df : pd.DataFrame
        Hourly load of some type. Must have the column "load_mw"
    utc_offset : int, optional
        Number of hours to shift away from the data timezone (e.g. UTC), by default 0

    Returns
    -------
    pd.DataFrame
        Time-shifted demand
    """
    df["load_mw"] = np.roll(df["load_mw"], utc_offset)
    return df


def state_demand_to_region(df: pd.DataFrame, pop: pd.DataFrame) -> pd.DataFrame:
    """Allocate hourly demand from states to regions based on population proportion

    Parameters
    ----------
    df : pd.DataFrame
        Hourly demand by one or more resource types in different states. Must have columns
        "state" and "time_index"
    pop : pd.DataFrame
        Proportion of state population in each region. Must have columns "state", "state_prop",
        and "region".


    Returns
    -------
    pd.DataFrame
        Hourly demand data ("load_mw") grouped by "time_index" and "region"
    """
    temp = pd.merge(df, pop, on=["state"], how="left")
    temp["load_mw"] *= temp["state_prop"]
    temp = temp.groupby(["time_index", "region"], as_index=False)["load_mw"].sum()
    return temp


@deep_freeze_args
@lru_cache()
def electrification_profiles(
    stock_fn: str,
    year: int,
    elec_scenario: str,
    regions: List[str],
    utc_offset: int = 0,
    path_in: Path = None,
) -> pd.DataFrame:
    """Create demand profiles for potentially flexible resources that will be electrified.

    Profiles are shifted from stored timezone (assume UTC) to the model timezone.

    Parameters
    ----------
    stock_fn : str
        Name of the data file with stock values for each year.
    year : int
        Planning period or model year. Used to select stock values for flexible resources
        and their demand profiles.
    elec_scenario : str
        Name of a scenario from the stock data file.
    regions : List[str]
        All of the base regions that will have profiles created.
    utc_offset : int, optional
        Hours to shift data from UTC to the desired timezone for the model, by default 0.
    path_in : Path, optional
        Folder where stock and incremental factor (profile) data are located, by default
        None.

    Returns
    -------
    pd.DataFrame
        Columns include "time_index", "region", "load_mw", and "resource".

    Raises
    ------
    KeyError
        No "EFS_DATA" key is included in the .env file.
    FileNotFoundError
        The "EFS_DATA" or "path_in" folder does not exist.
    ValueError
        The "electrificication_scenario" parameter does not match scenarios in the stock
        file.
    """
    logger.info("Building demand profiles of electrified end-use technologies")
    if not path_in:
        try:
            path_in = Path(SETTINGS["EFS_DATA"])
        except TypeError:
            raise KeyError("The variable 'EFS_DATA' is not included in your .env file.")
    if not path_in.is_dir():
        raise FileNotFoundError(
            f"The folder with EFS/flexible demand data ({str(path_in)}) was not found."
        )

    pop_files = path_in.glob("*pop_weight*")
    newest_pop_file = max(pop_files, key=os.path.getctime)
    pop = load_region_pop_frac(path_in=path_in, fn=newest_pop_file.name)
    pop = pop.loc[pop["region"].isin(regions), :]
    states = pop["state"].unique()
    scenario_stock = pd.read_parquet(path_in / stock_fn)
    scenario_stock.columns = snake_case_col(scenario_stock.columns)
    valid_scenarios = list(scenario_stock.scenario.unique())
    if elec_scenario not in valid_scenarios:
        raise ValueError(
            f"Your 'electrification_scenario' parameter value '{elec_scenario}' was not "
            f"found in the 'electrification_stock_fn' file '{stock_fn}'. Valid scenarios "
            f"are: {valid_scenarios}."
        )
    scenario_stock = scenario_stock.loc[
        (scenario_stock["year"] == year)
        & (scenario_stock["scenario"] == elec_scenario)
        & (scenario_stock["state"].isin(states)),
        :,
    ]

    subsector_ts_dfs = []
    for name, (sector, subsector) in running_sectors.items():
        df = create_subsector_ts(
            sector, subsector, year, scenario_stock, utc_offset, path_in
        ).pipe(state_demand_to_region, pop)

        df["resource"] = name
        subsector_ts_dfs.append(df)

    elec_profiles = pd.concat(subsector_ts_dfs)

    return elec_profiles
